import numpy as np
from cuqi.distribution import Distribution
from cuqi.distribution import Gaussian

class UserDefinedDistribution(Distribution):
    """
    Class to wrap user-defined logpdf, gradient, and/or sampling callable into CUQIpy Distribution.

    Parameters
    ------------
    logpdf_func: Function evaluating log probability density function. Callable.
    gradient_func: Function evaluating the gradient of the logpdf. Callable.
    sample_func: Function drawing samples from distribution. Callable.
    
    Methods
    -----------
    sample: generate one or more random samples
    logpdf: evaluate log probability density function
    gradient: evaluate gradient of logpdf
    
    Example
    -----------
    .. code-block:: python

        # Generate an i.i.d. n-dim Gaussian with zero mean and 2 variance.
        mu1 = -1.0
        std1 = 4.0
        X = cuqi.distribution.Normal(mean=mu1, std=std1)
        dim1 = 1
        logpdf_func = lambda xx: -np.log(std1*np.sqrt(2*np.pi))-0.5*((xx-mu1)/std1)**2
        sample_func = lambda : mu1 + std1*np.random.randn(dim1,1)
        XU = cuqi.distribution.UserDefinedDistribution(dim=dim1, logpdf_func=logpdf_func, sample_func=sample_func)
    """

    def __init__(self, dim=None, logpdf_func=None, gradient_func=None, sample_func=None, **kwargs):

        # Init from abstract distribution class
        super().__init__(**kwargs)

        if logpdf_func is not None and not callable(logpdf_func): raise ValueError("logpdf_func should be callable.")
        if sample_func is not None and not callable(sample_func): raise ValueError("sample_func should be callable.")
        if gradient_func is not None and not callable(gradient_func): raise ValueError("grad_func should be callable.")
        
        self.geometry = dim # Store dimension by calling setter for geometry (creates default geometry)
        self.logpdf_func = logpdf_func
        self.sample_func = sample_func
        self.gradient_func = gradient_func

    def logpdf(self, x):
        if self.logpdf_func is not None:
            return self.logpdf_func(x)
        else:
            raise Exception("logpdf_func is not defined.")
    
    def _gradient(self, x):
        if self.gradient_func is not None:
            return self.gradient_func(x)
        else:
            raise Exception("gradient_func is not defined.")

    def _sample(self, N=1, rng=None):
        #TODO(nabr) allow sampling more than 1 sample and potentially rng?
        if self.sample_func is not None:
            if N==1:
                return self.sample_func().flatten()
            else:
                out = np.zeros((self.dim,N))
                for i in range(N):
                    out[:,i] = self.sample_func()
                return out
        else:
            raise Exception("sample_func is not defined. Sampling can be performed by passing the density to a sampler from the 'sampler' module.")

    @property
    def _mutable_vars(self):
        """ Returns the mutable variables of the distribution. """
        # Currently mutable variables are not supported for user-defined distributions.
        return []

    def get_conditioning_variables(self):
        """ Returns the conditioning variables of the distribution. """
        # Currently conditioning variables are not supported for user-defined distributions.
        return []

### BENCHMARKS
class DistributionGallery(UserDefinedDistribution):
    # we follow the benchmarks from:
    # https://github.com/chi-feng/mcmc-demo/blob/master/main/MCMC.js

    def __init__(self, distribution_name,**kwargs):
        # Init from abstract distribution class
        if distribution_name == "CalSom91":
            #TODO: user can specify sig and delta
            dim = 2
            self.sig = 0.1
            self.delta = 1
            self.dfdx1 = lambda x1, x2: -(x1*(np.sqrt(x2**2+x1**2)-1))/(self.sig**2*np.sqrt(x2**2+x1**2))
            self.dfdx2 = lambda x1, x2: -(x2*(np.sqrt(x2**2+x1**2)-1))/(self.sig**2*np.sqrt(x2**2+x1**2))-(x2-1)/self.delta**2
            logpdf_func = self._CalSom91_logpdf_func
            grad_logpdf = self._CalSom91_grad_logpdf
        elif distribution_name == "BivariateGaussian":
            #TODO: user can specify Gaussain input
            #TODO: Keep Gaussian distribution other functionalities (e.g. _sample)
            dim = 2
            mu = np.zeros(dim)
            sigma = np.diag(np.linspace(0.5, 1, dim))
            R = np.array([[1.0, 0.9 ],[0.9, 1.0]])
            dist = Gaussian(mu, sigma@R@sigma)
            self._sample = dist._sample
            logpdf_func = dist.logpdf
            grad_logpdf = dist.gradient
        elif distribution_name == "funnel":
            # "funnel" distribution from Neal (2003) - Slice Sampling. Annals of Statistics 31(3): 705-67
            dim = 2
            self.m0, self.m1 = 0, 0
            self.s1 = 3
            self.f = lambda x, m, s: -0.5*np.log(2*np.pi) - np.log(s) - 0.5*((x-m)/s)**2
            self.dfdx = lambda x, m, s: -(x-m)/(s**2)
            self.dfds = lambda x, m, s: -1/s + ((x-m)**2)/(s**3)
            #
            logpdf_func = self._funnel_logpdf_func
            grad_logpdf = self._funnel_grad_logpdf
        elif distribution_name == "mixture":
            dim = 2
            self.m0, self.m1, self.m2 = np.array([-1.5, -1.5]), np.array([1.5, 1.5]), np.array([-2, 2])
            self.S0, self.S1, self.S2 = (0.8**2), (0.8**2), (0.5**2)
            self.G0 = Gaussian(self.m0, self.S0)
            self.G1 = Gaussian(self.m1, self.S1)
            self.G2 = Gaussian(self.m2, self.S2)
            # self.grad_logG = lambda x, mu, s: -np.diag([1/s, 1/s])@(x-mu).T
            #
            logpdf_func = self._mixture_logpdf_func
            grad_logpdf = self._mixture_grad_func
        elif distribution_name == "squiggle":
            dim = 2
            m0 = np.zeros(dim)
            S0 = np.array([[2, 0.25], [0.25, 0.5]])
            self.G0 = Gaussian(m0, S0)
            #
            logpdf_func = self._squiggle_logpdf_func
            grad_logpdf = self._squiggle_grad_logpdf
        elif distribution_name == "donut":
            dim = 2
            self.radius, self.sigma2 = 2.6, 0.033
            #
            logpdf_func = self._donut_logpdf_func
            grad_logpdf = self._donut_grad_logpdf
        elif distribution_name == "banana":
            dim = 2
            m0 = np.array([0, 4])
            S0 = np.array([[1, 0.5], [0.5, 1]])
            self.G0 = Gaussian(m0, S0)
            self.a, self.b = 2, 0.2
            #
            logpdf_func = self._banana_logpdf_func
            grad_logpdf = self._banana_grad_logpdf

        super().__init__(logpdf_func=logpdf_func, gradient_func=grad_logpdf, dim=dim, **kwargs)

    def _CalSom91_logpdf_func(self, x):
        if len(x.shape) == 1:
            x = x.reshape((1, 2))
        return -1/(2*self.sig**2)*(np.sqrt(x[:,0]**2+ x[:,1]**2) -1 )**2 -1/(2*self.delta**2)*(x[:,1]-1)**2

    def _CalSom91_grad_logpdf(self, x):
        if len(x.shape) == 1:
            x = x.reshape((1, 2))
        grad = np.array([self.dfdx1(x[:, 0], x[:, 1]), self.dfdx2(x[:, 0], x[:, 1])])
        if x.shape[0] == 1:
            return grad.flatten()
        else:
            return grad

    def _funnel_logpdf_func(self, x):
        if len(x.shape) == 1:
            x = x.reshape((1, 2))
        s0 = np.exp(x[:, 1]/2)
        return self.f(x[:, 0], self.m0, s0) + self.f(x[:, 1], self.m1, self.s1)

    def _funnel_grad_logpdf(self, x):
        if len(x.shape) == 1:
            x = x.reshape((1, 2))
        s0 = np.exp(x[:, 1]/2)
        grad = np.array([self.dfdx(x[:, 0], self.m0, s0), \
                        self.dfds(x[:, 0], self.m0, s0)*0.5*s0 + self.dfdx(x[:, 1], self.m1, self.s1)])
        if x.shape[0] == 1:
            return grad.flatten()
        else:
            return grad

    def _mixture_logpdf_func(self, x):
        if len(x.shape) == 1:
            x = x.reshape((1, 2))
        return np.log(self.G0.pdf(x) + self.G1.pdf(x) + self.G2.pdf(x))
    
    def _mixture_grad_func(self, x):
        if len(x.shape) == 1:
            x = x.reshape((1, 2))
        p1, p2, p3 = self.G0.pdf(x), self.G1.pdf(x), self.G2.pdf(x)
        scale = np.nan_to_num(1 / (p1 + p2 + p3))
        grad = (p1*self.G0.gradient(x) + p2*self.G1.gradient(x) + p3*self.G2.gradient(x)) * scale
        if x.shape[0] == 1:
            return grad.flatten()
        else:
            return grad

    def _squiggle_logpdf_func(self, x):
        if len(x.shape) == 1:
            x = x.reshape((1, 2))
        y = np.zeros((x.shape[0], self.dim))
        y[:, 0], y[:, 1] = x[:, 0], x[:, 1] + np.sin(5*x[:, 0])
        return self.G0.logpdf(y)

    def _squiggle_grad_logpdf(self, x):
        if len(x.shape) == 1:
            x = x.reshape((1, 2))
        y = np.zeros((x.shape[0], self.dim))
        y[:, 0], y[:, 1] = x[:, 0], x[:, 1] + np.sin(5*x[:, 0])
        grad = self.G0.gradient(y)
        gradx0, gradx1 = grad[0, :] + grad[1, :]*5*np.cos(5*x[:, 0]), grad[1, :]
        grad[0, :], grad[1, :] = gradx0, gradx1
        if x.shape[0] == 1:
            return grad.flatten()
        else:
            return grad

    def _donut_logpdf_func(self, x):
        if len(x.shape) == 1:
            x = x.reshape((1, 2))
        r = np.linalg.norm(x, axis=1)
        return - (r - self.radius)**2 / self.sigma2

    def _donut_grad_logpdf(self, x):
        if len(x.shape) == 1:
            x = x.reshape((1, 2))
        r = np.linalg.norm(x, axis=1)
        idx = np.argwhere(r==0)
        r[idx] = 1e-16
        grad = np.array([(x[:, 0]*((self.radius/r)-1)*2)/self.sigma2, \
                         (x[:, 1]*((self.radius/r)-1)*2)/self.sigma2])
        if x.shape[0] == 1:
            return grad.flatten()
        else:
            return grad

    def _banana_logpdf_func(self, x):
        if len(x.shape) == 1:
            x = x.reshape((1, 2))
        y = np.zeros((x.shape[0], self.dim))
        y[:, 0] = x[:, 0]/self.a
        y[:, 1] = x[:, 1]*self.a + self.a*self.b*(x[:, 0]**2 + self.a**2)
        return self.G0.logpdf(y)

    def _banana_grad_logpdf(self, x):
        if len(x.shape) == 1:
            x = x.reshape((1, 2))
        y = np.zeros((x.shape[0], self.dim))
        y[:, 0] = x[:, 0]/self.a
        y[:, 1] = x[:, 1]*self.a + self.a*self.b*(x[:, 0]**2 + self.a**2)
        grad = self.G0.gradient(y)
        gradx0, gradx1 = grad[0, :]/self.a + grad[1, :]*self.a*self.b*2*x[:, 0], grad[1, :]*self.a
        grad[0, :], grad[1, :] = gradx0, gradx1
        if x.shape[0] == 1:
            return grad.flatten()
        else:
            return grad



from cuqi.geometry import _DefaultGeometry, _get_identity_geometries
from cuqi.distribution import Distribution

# ========================================================================
class Posterior(Distribution):
    """
    Posterior probability distribution defined by likelihood and prior.
    The geometry is automatically determined from the model and prior.
    Generates instance of cuqi.distribution.Posterior
    
    Parameters
    ------------
    likelihood: Likelihood function, cuqi.likelihood.Likelihood.
    prior: Prior distribution, cuqi.distribution.Distribution.

    """
    def __init__(self, likelihood, prior, **kwargs):

        if len(likelihood.get_parameter_names()) > 1:
            raise ValueError("Likelihood must only have one parameter.")
        if prior.is_cond:
            raise ValueError("Prior must not be a conditional distribution.")

        self.likelihood = likelihood
        self.prior = prior 
        super().__init__(**kwargs)

    @property
    def data(self):
        return self.likelihood.data

    @property
    def dim(self):
        return self.prior.dim

    @property
    def geometry(self):
        return self._geometry

    @geometry.setter
    def geometry(self, value):
        # Compare model and prior
        if self.model is not None and self.model.domain_geometry != self.prior.geometry:
            if isinstance(self.prior.geometry, _DefaultGeometry):
                pass #We allow default geometry in prior
            else:
                raise ValueError("Geometry from likelihood (model.domain_geometry) does not match prior geometry")

        # Compare value and prior
        if self.model is None and value is not None and value != self.prior.geometry:
            if isinstance(self.prior.geometry, _DefaultGeometry):
                pass #We allow default geometry in prior
            else:
                raise ValueError("Posterior and prior geometries are inconsistent.")

        # Compare model and value
        if self.model is not None and value is not None and value != self.model.domain_geometry:
            if isinstance(self.model.domain_geometry, _DefaultGeometry):
                pass #Allow default model geometry
            else:
                raise ValueError("Set geometry does not match with model geometry.")
            
        # Compare likelihood and prior
        if self.likelihood.geometry != self.prior.geometry:
            if isinstance(self.prior.geometry, _DefaultGeometry):
                pass #We allow default geometry in prior
            elif isinstance(self.likelihood.geometry, _DefaultGeometry):
                pass #We allow default geometry in likelihood
            else:
                raise ValueError("Likelihood and prior geometries are inconsistent.")
        
        # If value is set, its consistant with prior (and prior is consistant with model)
        # Likelihood and prior are consistant.
        # If value is not set, take from model (if exists) or from likelihood or prior as last resort
        if value is not None and not isinstance(value, _DefaultGeometry):
            self._geometry = value
        elif self.model is not None and not isinstance(self.model.domain_geometry, _DefaultGeometry):
            self._geometry = self.model.domain_geometry
        elif not isinstance(self.likelihood.geometry, _DefaultGeometry):
            self._geometry = self.likelihood.geometry
        else:
            self._geometry = self.prior.geometry
            
    def logpdf(self, *args, **kwargs):
        """ Returns the logpdf of the posterior distribution"""
        return self.likelihood.logd(*args, **kwargs)+ self.prior.logd(*args, **kwargs)

    def get_conditioning_variables(self):
        return self.prior.get_conditioning_variables()

    def get_parameter_names(self):
        return self.prior.get_parameter_names()

    def _gradient(self, x):
        #Avoid complicated geometries that change the gradient.
        if not type(self.geometry) in _get_identity_geometries() and\
           not hasattr(self.geometry, 'gradient'):
            raise NotImplementedError("Gradient not implemented for distribution {} with geometry {}".format(self,self.geometry))
            
        return self.likelihood.gradient(x)+ self.prior.gradient(x)        

    def _sample(self,N=1,rng=None):
        raise Exception("'Posterior.sample' is not defined. Sampling can be performed with the 'sampler' module.")

    @property
    def model(self):
        """Extract the cuqi model from likelihood."""
        return self.likelihood.model

    def __repr__(self):
        msg = f"Posterior(\n"
        msg += "    Equation:\n"
        msg += f"\t p({self.prior.name}|{self.likelihood.name}) ∝ L({self.prior.name}|{self.likelihood.name})p({self.prior.name})\n"
        msg += "    Densities:\n"
        msg += f"\t{self.likelihood.name} ~ {self.likelihood}\n "
        msg += f"\t{self.prior.name} ~ {self.prior}\n "
        msg += ")"
        return msg



import numpy as np
from cuqi.distribution import Distribution

class Uniform(Distribution):


    def __init__(self, low=None, high=None, is_symmetric=True, **kwargs):
        """
        Parameters
        ----------
        low : float or array_like of floats
            Lower bound(s) of the uniform distribution.
        high : float or array_like of floats 
            Upper bound(s) of the uniform distribution.
        """
        # Init from abstract distribution class
        super().__init__(is_symmetric=is_symmetric, **kwargs)       

        # Init specific to this distribution
        self.low = low
        self.high = high      

    def logpdf(self, x):
        # First check whether x is outside bounds.
        # It is outside if any coordinate is outside the interval.
        if np.any(x < self.low) or np.any(x > self.high):
            # If outside always return -inf
            return_val = -np.inf  
        else:
            # If inside, compute the area and obtain the constant 
            # probability (pdf) as 1 divided by the area, the convert 
            # to logpdf. Special case if scalar.
            diff = self.high - self.low
            if isinstance(diff, (list, tuple, np.ndarray)): 
                v= np.prod(diff)
            else:
                v = diff
            return_val = np.log(1.0/v)
        return return_val

    def _sample(self,N=1, rng=None):

        if rng is not None:
            s = rng.uniform(self.low, self.high, (N,self.dim)).T
        else:
            s = np.random.uniform(self.low, self.high, (N,self.dim)).T

        return s


import numpy as np
import cuqi
from cuqi.sampler import ProposalBasedSampler


class MH(ProposalBasedSampler):
    """Metropolis Hastings sampler.

    Allows sampling of a target distribution by random-walk sampling of a proposal distribution along with an accept/reject step.

    Parameters
    ----------

    target : `cuqi.distribution.Distribution` or lambda function
        The target distribution to sample. Custom logpdfs are supported by using a :class:`cuqi.distribution.UserDefinedDistribution`.
    
    proposal : `cuqi.distribution.Distribution` or callable method
        The proposal to sample from. If a callable method it should provide a single independent sample from proposal distribution. Defaults to a Gaussian proposal.  *Optional*.

    scale : float
        Scale parameter used to define correlation between previous and proposed sample in random-walk.  *Optional*.

    x0 : ndarray
        Initial parameters. *Optional*

    dim : int
        Dimension of parameter space. Required if target and proposal are callable functions. *Optional*.

    callback : callable, *Optional*
        If set this function will be called after every sample.
        The signature of the callback function is `callback(sample, sample_index)`,
        where `sample` is the current sample and `sample_index` is the index of the sample.
        An example is shown in demos/demo31_callback.py.

    Example
    -------
    .. code-block:: python

        # Parameters
        dim = 5 # Dimension of distribution
        mu = np.arange(dim) # Mean of Gaussian
        std = 1 # standard deviation of Gaussian

        # Logpdf function
        logpdf_func = lambda x: -1/(std**2)*np.sum((x-mu)**2)

        # Define distribution from logpdf as UserDefinedDistribution (sample and gradients also supported)
        target = cuqi.distribution.UserDefinedDistribution(dim=dim, logpdf_func=logpdf_func)

        # Set up sampler
        sampler = cuqi.sampler.MH(target, scale=1)

        # Sample
        samples = sampler.sample(2000)

    """
    #target,  proposal=None, scale=1, x0=None, dim=None
    #    super().__init__(target, proposal=proposal, scale=scale,  x0=x0, dim=dim)
    def __init__(self, target, proposal=None, scale=None, x0=None, dim=None, **kwargs):
        """ Metropolis-Hastings (MH) sampler. Default (if proposal is None) is random walk MH with proposal that is Gaussian with identity covariance"""
        super().__init__(target, proposal=proposal, scale=scale,  x0=x0, dim=dim, **kwargs)


    @ProposalBasedSampler.proposal.setter 
    def proposal(self, value):
        fail_msg = "Proposal should be either None, symmetric cuqi.distribution.Distribution or a lambda function."

        if value is None:
            self._proposal = cuqi.distribution.Gaussian(np.zeros(self.dim), 1)
        elif not isinstance(value, cuqi.distribution.Distribution) and callable(value):
            raise NotImplementedError(fail_msg)
        elif isinstance(value, cuqi.distribution.Distribution) and value.is_symmetric:
            self._proposal = value
        else:
            raise ValueError(fail_msg)
        self._proposal.geometry = self.target.geometry

    def _sample(self, N, Nb):
        if self.scale is None:
            raise ValueError("Scale must be set to sample without adaptation. Consider using sample_adapt instead.")
        
        Ns = N+Nb   # number of simulations

        # allocation
        samples = np.empty((self.dim, Ns))
        target_eval = np.empty(Ns)
        acc = np.zeros(Ns, dtype=int)

        # initial state    
        samples[:, 0] = self.x0
        target_eval[0] = self.target.logd(self.x0)
        acc[0] = 1

        # run MCMC
        for s in range(Ns-1):
            # run component by component
            samples[:, s+1], target_eval[s+1], acc[s+1] = self.single_update(samples[:, s], target_eval[s])
            self._print_progress(s+2,Ns) #s+2 is the sample number, s+1 is index assuming x0 is the first sample
            self._call_callback(samples[:, s+1], s+1)

        # remove burn-in
        samples = samples[:, Nb:]
        target_eval = target_eval[Nb:]
        accave = acc[Nb:].mean()   
        print('\nAverage acceptance rate:', accave, '\n')
        #
        return samples, target_eval, accave

    def _sample_adapt(self, N, Nb):
        # Set intial scale if not set
        if self.scale is None:
            self.scale = 0.1
            
        Ns = N+Nb   # number of simulations

        # allocation
        samples = np.empty((self.dim, Ns))
        target_eval = np.empty(Ns)
        acc = np.zeros(Ns)

        # initial state    
        samples[:, 0] = self.x0
        target_eval[0] = self.target.logd(self.x0)
        acc[0] = 1

        # initial adaptation params 
        Na = int(0.1*N)                              # iterations to adapt
        hat_acc = np.empty(int(np.floor(Ns/Na)))     # average acceptance rate of the chains
        lambd = self.scale
        star_acc = 0.234    # target acceptance rate RW
        i, idx = 0, 0

        # run MCMC
        for s in range(Ns-1):
            # run component by component
            samples[:, s+1], target_eval[s+1], acc[s+1] = self.single_update(samples[:, s], target_eval[s])
            
            # adapt prop spread using acc of past samples
            if ((s+1) % Na == 0):
                # evaluate average acceptance rate
                hat_acc[i] = np.mean(acc[idx:idx+Na])

                # d. compute new scaling parameter
                zeta = 1/np.sqrt(i+1)   # ensures that the variation of lambda(i) vanishes
                lambd = np.exp(np.log(lambd) + zeta*(hat_acc[i]-star_acc))

                # update parameters
                self.scale = min(lambd, 1)

                # update counters
                i += 1
                idx += Na

            # display iterations
            self._print_progress(s+2,Ns) #s+2 is the sample number, s+1 is index assuming x0 is the first sample


        # remove burn-in
        samples = samples[:, Nb:]
        target_eval = target_eval[Nb:]
        accave = acc[Nb:].mean()   
        print('\nAverage acceptance rate:', accave, 'MCMC scale:', self.scale, '\n')
        
        return samples, target_eval, accave


    def single_update(self, x_t, target_eval_t):
        # propose state
        xi = self.proposal.sample(1)   # sample from the proposal
        x_star = x_t + self.scale*xi.flatten()   # MH proposal

        # evaluate target
        target_eval_star = self.target.logd(x_star)

        # ratio and acceptance probability
        ratio = target_eval_star - target_eval_t  # proposal is symmetric
        alpha = min(0, ratio)

        # accept/reject
        u_theta = np.log(np.random.rand())
        if (u_theta <= alpha):
            x_next = x_star
            target_eval_next = target_eval_star
            acc = 1
        else:
            x_next = x_t
            target_eval_next = target_eval_t
            acc = 0
        
        return x_next, target_eval_next, acc



import cuqi
import numpy as np
import pytest

def test_joint_dist_dim_geometry():
    """ Test the dimension and geometry properties of a joint distribution """

    model, data, _ = cuqi.testproblem.Poisson1D().get_components() # Model is m times n, m != n.

    # Bayesian model
    d = cuqi.distribution.Gamma(1, 1e-4)
    x = cuqi.distribution.Gaussian(np.zeros(model.domain_dim), lambda d: d)
    y = cuqi.distribution.Gaussian(model, 1)

    # Joint distribution
    J = cuqi.distribution.JointDistribution(d, x, y)

    # Test the dimensions
    assert J.dim == [d.dim, x.dim, y.dim]

    # Test the geometries
    assert J.geometry == [d.geometry, x.geometry, y.geometry]

    # Test dim and geometry for posterior
    P = J(y=data)

    assert P.dim == [d.dim, x.dim]

    assert P.geometry == [d.geometry, x.geometry]


@pytest.mark.parametrize("densities", [
    [
        cuqi.distribution.Gamma(1, 1e-4, name="x"),
        cuqi.distribution.Normal(0, 1, name="y")
    ],
    [
        cuqi.distribution.Gamma(1, 1e-4, name="x"),
        cuqi.distribution.Normal(0, lambda x:x, name="y")
    ],
    [
        cuqi.distribution.Gamma(1, 1e-4, name="d"),
        cuqi.distribution.Gamma(1, 1e-2, name="l"),
        cuqi.distribution.Gaussian(np.zeros(8), lambda d: d, name="x"),
        cuqi.distribution.Gaussian(
            mean=cuqi.testproblem.Deconvolution1D(dim=8).model,
            cov=lambda l: l,
            name="y"
        )
    ],
    [
        cuqi.distribution.Normal(0, 1e-2, name="z"),
        cuqi.distribution.Gamma(1, lambda z: abs(z), name="d"),
        cuqi.distribution.Gamma(lambda z: z, 1e-2, name="l"),
        cuqi.distribution.Gaussian(np.zeros(8), lambda d: d, name="x"),
        cuqi.distribution.Gaussian(
            mean=cuqi.testproblem.Deconvolution1D(dim=8).model,
            cov=lambda l: l,
            name="y"
        )
    ],
])
def test_joint_dist_logd(densities):
    """ Tests the log density function of the Joint distribution """

    # Create a joint distribution
    J = cuqi.distribution.JointDistribution(*densities)

    # Get the parameter names
    names = J.get_parameter_names()

    # Get dimensions
    dims = J.dim

    # Make dictionary of random parameters matching names and dimensions
    params = {name: abs(np.random.randn(dim)) for name, dim in zip(names, dims)}

    # Evaluate the log density function at random points
    logd = J.logd(**params)

    # Manually evaluate the log density function for each density
    logd_manual = 0
    for density in densities:
        # Extract the parameters for this density
        density_params = {name: params[name] for name in density.get_parameter_names()}
        logd_manual += density.logd(**density_params)

    # Compare the log density functions
    assert logd == pytest.approx(logd_manual)

@pytest.mark.parametrize("densities", [
    [
        cuqi.distribution.Gamma(1, 1e-4, name="x"),
        cuqi.distribution.Gaussian(0, lambda x:x, name="z"),
        cuqi.distribution.Normal(0, lambda x:x, name="y")
    ],
    [
        cuqi.distribution.Gamma(1, 1e-4, name="d"),
        cuqi.distribution.Gamma(1, 1e-2, name="l"),
        cuqi.distribution.Gaussian(np.zeros(8), lambda d: d, name="x"),
        cuqi.distribution.Gaussian(
            mean=cuqi.testproblem.Deconvolution1D(dim=8).model,
            cov=lambda l: l,
            name="y"
        )
    ],
    [
        cuqi.distribution.Normal(0, 1e-2, name="z"),
        cuqi.distribution.Gamma(1, lambda z: abs(z), name="d"),
        cuqi.distribution.Gamma(lambda z: z, 1e-2, name="l"),
        cuqi.distribution.Gaussian(np.zeros(8), lambda d: d, name="x"),
        cuqi.distribution.Gaussian(
            mean=cuqi.testproblem.Deconvolution1D(dim=8).model,
            cov=lambda l: l,
            name="y"
        )
    ],
])
def test_joint_dist_properties(densities):
    """ This tests various properties of the joint distribution """

    # Create a joint distribution
    J = cuqi.distribution.JointDistribution(*densities)

    # Now check various properties. First check the dimension
    assert J.dim == [density.dim for density in densities]

    # Check the geometry
    assert J.geometry == [density.geometry for density in densities]

    # Check the parameter names
    assert J.get_parameter_names() == [density.name for density in densities]

    # Check list of distributions
    assert J._distributions == densities

    # Check likelihoods
    assert J._likelihoods == []

    # Now we condition y on some data
    data = cuqi.testproblem.Deconvolution1D(dim=8).data

    P = J(y=data)

    # Check the dimension
    assert P.dim == [density.dim for density in densities[:-1]]

    # Check the geometry
    assert P.geometry == [density.geometry for density in densities[:-1]]

    # Check the parameter names
    assert P.get_parameter_names() == [density.name for density in densities[:-1]]

    # Check list of distributions by comparing the names
    assert [dist.name for dist in P._distributions] == [density.name for density in densities[:-1]]

    # Check likelihoods by comparing the names
    assert [L.name for L in P._likelihoods] == [densities[-1].name]

def test_joint_dist_reduce():
    """ This tests the reduce hack for the joint distribution. """

    J, data = hierarchical_joint()   

    # Check if we get the expected result when conditioning.
    assert isinstance(J(y=data), cuqi.distribution.JointDistribution)
    assert isinstance(J(y=data, z=1, d=1, l=1), cuqi.distribution.Posterior)
    assert isinstance(J(y=data, z=1, d=1, x=np.zeros(8)), cuqi.distribution.Posterior)
    assert isinstance(J(y=data, z=1, x=np.zeros(8), l=1), cuqi.distribution.Posterior)
    assert isinstance(J(y=data, d=1, x=np.zeros(8), l=1), cuqi.distribution.JointDistribution) # 2 likelihoods

def test_stacked_joint_distribution():
    """ This tests the stacked joint distribution """

    J, data = hierarchical_joint()

    Js = J._as_stacked()

    # Check the dimension
    assert Js.dim == sum(J.dim)

    # Check the geometry
    assert Js.geometry == cuqi.geometry._DefaultGeometry1D(Js.dim)

    # Check you can evaluate the log density function with a single vector
    Js.logd(np.ones(Js.dim))

    # Check that you can condition on the stacked distribution
    # and again evaluate as a single vector
    Ps = Js(y=data)
    Ps.logd(np.ones(Ps.dim))

    # Even if you condition in multiple parameters
    Ps2 = Js(y=data, z=1, d=1)
    Ps2.logd(np.ones(Ps2.dim))

    # And even if you condition on all parameters
    Ps3 = Js(y=data, z=1, d=1, x=np.zeros(Js.get_density("x").dim), l=1)
    #Ps3.logd() # Should be allowed with no arguments

def hierarchical_joint(main_dim=8):
    """ This creates a deconvolution based hierarchical joint distribution.
    
    Parameters
    ----------
    main_dim : int
        The dimension of the main model.

    Returns
    -------
    JointDistribution, data
        
    """

    densities = [
        cuqi.distribution.Normal(0, 1e-2, name="z"),
        cuqi.distribution.Gamma(1, lambda z: abs(z), name="d"),
        cuqi.distribution.Gamma(lambda z: z, 1e-2, name="l"),
        cuqi.distribution.Gaussian(np.zeros(main_dim), lambda d: d, name="x"),
        cuqi.distribution.Gaussian(
            mean=cuqi.testproblem.Deconvolution1D(dim=main_dim).model,
            cov=lambda l: l,
            name="y"
        )
    ]

    data = cuqi.testproblem.Deconvolution1D(dim=main_dim).data

    J = cuqi.distribution.JointDistribution(*densities)

    return J, data

def test_extra_parameter_no_prior():
    A = np.random.randn(10,3)

    # Define distributions for Bayesian model
    y = cuqi.distribution.Normal(lambda x: A@x, np.ones(10))
    x = cuqi.distribution.Normal(np.zeros(3), lambda z,b:z+b)
    z = cuqi.distribution.Gamma(1, 1)

    # Joint distribution p(y,x,z)
    with pytest.raises(ValueError, match=r"Missing prior for \['b'\]"):
        cuqi.distribution.JointDistribution(y,x,z)


def test_JointDistribution_reduce_MultipleLikelihoodPosterior():
    """ This tests if the joint distribution can be reduced to MultipleLikelihoodPosterior """

    J, data = hierarchical_joint()

    # Posterior w.r.t z is with multiple likelihoods if all other parameters are fixed
    # for this J
    posterior_z = J(y=data, d=0.1, l=0.1, x=data)

    # Check we get the right class
    assert isinstance(posterior_z, cuqi.distribution.MultipleLikelihoodPosterior)

    # Check the dimension
    assert posterior_z.dim == 1

    # Check the "prior" in multiple likelihood posterior matches z
    assert posterior_z.prior.mean == J.get_density("z").mean
    assert posterior_z.prior.std == J.get_density("z").std

    # Check logd can be evaluated (tolerance is high since new data every time)
    assert np.allclose(posterior_z.logd(1), -5000, atol=10)

def test_MultipleLikelihoodPosterior_should_raise_if_two_densities():
    """ This tests if the MultipleLikelihoodPosterior raises if it has two densities. """

    # Define distributions for Bayesian model
    y = cuqi.distribution.Normal(lambda x: x, 1)
    x = cuqi.distribution.Normal(0, 1)

    # Joint distribution p(y,x)
    J = cuqi.distribution.JointDistribution(y,x)

    # Posterior
    P = J(y=1)

    # Check we get the right class
    assert isinstance(P, cuqi.distribution.Posterior)

    # Check that we cannot create MultipleLikelihoodPosterior
    with pytest.raises(ValueError, match=r"requires at least three densities"):
        cuqi.distribution.MultipleLikelihoodPosterior(y.to_likelihood(1), x)


def test_MultipleLikelihoodPosterior_should_raise_if_no_likelihood():
    """ This tests if the MultipleLikelihoodPosterior raises if no likelihood is given."""

    # Define distributions for Bayesian model
    y1 = cuqi.distribution.Normal(lambda x: x, 1)
    y2 = cuqi.distribution.Normal(lambda x: x+1, 1)
    x = cuqi.distribution.Normal(0, 1)
 
    # Check that we cannot create MultipleLikelihoodPosterior
    with pytest.raises(ValueError, match=r"must have a likelihood and prior"):
        cuqi.distribution.MultipleLikelihoodPosterior(y1, y2, x)

def test_MultipleLikelihoodPosterior_should_raise_if_names_do_not_match():
    """ This tests if the MultipleLikelihoodPosterior raises if the names do not match."""

    # Define distributions for Bayesian model
    y = cuqi.distribution.Normal(lambda x: x, 1)
    x = cuqi.distribution.Normal(0, 1)
    z = cuqi.distribution.Normal(0, 1)
 
    # Check that we cannot create MultipleLikelihoodPosterior
    with pytest.raises(ValueError, match=r"the same parameter name"):
        cuqi.distribution.MultipleLikelihoodPosterior(y.to_likelihood(1), x, z)




import numpy as np
from scipy.sparse import spdiags, eye, kron, vstack

# ========== Operator class ===========
class Operator(object):
    """
    A linear operator which is represented by a matrix. 
    """

    def __init__(self):
        self._matrix = None
        pass

    def __matmul__(self, vec):
        return self._matrix@vec

    def __rmatmul__(self, vec):
        return vec@self._matrix

    def __add__(self, val):
        return self._matrix+val

    def __radd__(self, val):
        return self.__add__(val)

    def __mul__(self, val):
        return self._matrix*val

    def __rmul__(self, val):
        return self.__mul__(val)

    @property
    def T(self):
        return self._matrix.T

    @property
    def shape(self):
        return self._matrix.shape

    def get_matrix(self):
        return self._matrix

class FirstOrderFiniteDifference(Operator):
    """
    First order finite difference differential operator for 1D and 2D grids. 

    Attributes:
    -----------
        num_nodes: int or tuple
            For a 1D operator, num_nodes is a one dimensional tuple or an integer representing the number of discretization nodes in a 1D grid. For a 2D operator, num_nodes is a two dimensional tuple of integers representing the number of discretization nodes in the 2D grid in the x axis and the y axis, respectively.

        bc_type: str
            The boundary condition type for the operator. 

        physical_dim: int
            Either 1 or 2 for 1D and 2D operators, respectively.

        dx : int or float
            The grid spacing (length between two consecutive nodes). 
    """

    def __init__(self, num_nodes, bc_type= 'periodic', dx=None):	

        if isinstance(num_nodes, (int,np.integer)):
            self._num_nodes = (num_nodes,)

        elif isinstance(num_nodes, tuple) and len(num_nodes) in [1,2] and\
            np.all([isinstance(num,(int,np.integer)) for num in num_nodes]):
            
            if len(num_nodes)==2 and (num_nodes[0] != num_nodes[1]):
                raise NotImplementedError("The case in which the number of nodes in the x direction is not equal to the number of nodes in the y direction is not implemented.")
            self._num_nodes = num_nodes
        
        else:
            raise ValueError("num_nodes should be either and integer or a two dimensional tuple of integers")

        self._bc_type = bc_type
        if dx == None:
            self._dx = 1
        elif self.physical_dim == 1:
            self._dx = dx #TODO: check dx is a scalar value
        else:
            raise NotImplementedError('Specifying dx for a 2D operator is not implemented')
        self._create_diff_matrix()


    @property
    def num_nodes(self):
        return self._num_nodes

    @property
    def physical_dim(self):
        return len(self.num_nodes)

    @property
    def bc_type(self):
        return self._bc_type

    @property
    def dim(self):
        return np.prod(self.num_nodes)


    def _create_diff_matrix(self):
        if self.physical_dim == 2:
            assert(self.num_nodes[0] == self.num_nodes[1]), "The case in which self.num_nodes[0] != self.num_nodes[1] is not handled."
            N = self.num_nodes[0]
        elif self.physical_dim == 1:
            N = self.num_nodes[0]
        else:
            raise Exception("Cannot define N")

        # finite difference matrix
        one_vec = np.ones(N)
        diags = np.vstack([-one_vec, one_vec])
        if (self.bc_type == 'zero'):
            locs = [-1, 0]
            Dmat = spdiags(diags, locs, N+1, N)
        elif (self.bc_type == 'periodic'):
            locs = [-1, 0]
            Dmat = spdiags(diags, locs, N+1, N).tocsr()
            Dmat[-1, 0] = 1
            Dmat[0, -1] = -1
        elif (self.bc_type == 'neumann'):
            locs = [0, 1]
            Dmat = spdiags(diags, locs, N-1, N)
        elif (self.bc_type == 'backward'):
            locs = [0, -1]
            Dmat = spdiags(diags, locs, N, N).tocsr()
            Dmat[0, 0] = 1
        elif (self.bc_type == 'none'):
            Dmat = eye(N)
        else:
            raise ValueError(f"Unknown boundary type {self.bc_type}")

        # structure matrix
        if (self.physical_dim == 1):
            self._matrix = Dmat/self._dx

        elif (self.physical_dim == 2):            
            I = eye(N, dtype=int)
            Ds = kron(I, Dmat)
            Dt = kron(Dmat, I)
            self._matrix = vstack([Ds, Dt])
        
class SecondOrderFiniteDifference(FirstOrderFiniteDifference):
    """
    Second order finite difference differential operator for 1D and 2D grids. 

    Attributes:
    -----------
        num_nodes: int or tuple
            For a 1D operator, num_nodes is a one dimensional tuple or an integer representing
            the number of discretization nodes in a 1D grid. For a 2D operator, num_nodes is a
            two dimensional tuple of integers representing the number of discretization nodes
            in the 2D grid in the x axis and the y axis, respectively.

        bc_type: str
            The boundary condition type for the operator. 

        physical_dim: int
            Either 1 or 2 for 1D and 2D operators, respectively.

        dx : int or float
            The grid spacing (length between two consecutive nodes). 
    """

    def _create_diff_matrix(self):
        if self.physical_dim == 2:
            assert(self.num_nodes[0] == self.num_nodes[1]), "The case in which self.num_nodes[0] != self.num_nodes[1] is not handled."
            N = self.num_nodes[0]
        elif self.physical_dim == 1:
            N = self.num_nodes[0]
        else:
            raise Exception("Cannot define N")

        # finite difference matrix
        one_vec = np.ones(N)
        diags = np.vstack([-one_vec, 2*one_vec, -one_vec])
        if (self.bc_type == 'zero'):
            locs = [-2, -1, 0]
            Dmat = spdiags(diags, locs, N+2, N)
        elif (self.bc_type == 'periodic'):
            locs = [-2, -1, 0]
            Dmat = spdiags(diags, locs, N+2, N).tocsr()
            Dmat[0, -2] = -1
            Dmat[0:2, -1] = [2, -1]
            Dmat[-2, 0] = -1
            Dmat[-1, 0:2] = [2, -1]
        elif (self.bc_type == 'neumann'):
            locs = [0, 1, 2]
            Dmat = spdiags(diags, locs, N-2, N).tocsr()
        else:
            raise ValueError(f"Unknown boundary type {self.bc_type}")

        # structure matrix
        if (self.physical_dim == 1):
            self._matrix = Dmat/self._dx**2

        elif (self.physical_dim == 2):            
            I = eye(N, dtype=int)
            Ds = kron(I, Dmat)
            Dt = kron(Dmat, I)
            self._matrix = vstack([Ds, Dt])


class PrecisionFiniteDifference(Operator):
    """
    Precision operator constructed with a finite difference operator. 

    Attributes:
    -----------
        num_nodes: int or tuple
            For a 1D operator, num_nodes is a one dimensional tuple or an integer representing the number of discretization nodes in a 1D grid. For a 2D operator, num_nodes is a two dimensional tuple of integers representing the number of discretization nodes in the 2D grid in the x axis and the y axis, respectively.

        bc_type: str
            The boundary condition type for the finite difference operator. 

        physical_dim: int
            Either 1 or 2 for 1D and 2D operators, respectively. 
        
        order: int
            | The order of the finite difference operator.
            | Order 0: Identity operator.
            | Order 1: First order finite difference operator. 1D precision has a banded diagonal structure with [-1, 2, -1].
            | Order 2: Second order finite difference operator. 1D precision has a banded diagonal structure with [1, -4,  6, -4, 1].
    """
    def __init__(self, num_nodes , bc_type= 'periodic', order =1):
        if order == 0:
            self._diff_op = FirstOrderFiniteDifference(num_nodes, "none") # Special case that is idendity operator
        elif order == 1:
            self._diff_op = FirstOrderFiniteDifference(num_nodes, bc_type=bc_type)
        elif order == 2:
            self._diff_op = SecondOrderFiniteDifference(num_nodes, bc_type=bc_type)
        else:
            raise NotImplementedError
        self._create_prec_matrix()

    @property
    def physical_dim(self):
        return self._diff_op.physical_dim

    @property
    def num_nodes(self):
        return self._diff_op.num_nodes

    @property
    def bc_type(self):
        return self._diff_op.bc_type

    @property
    def dim(self):
        return self._diff_op.dim

    def _create_prec_matrix(self):
        if self.physical_dim == 1 or self.physical_dim == 2:
            self._matrix = (self._diff_op.T @ self._diff_op).tocsc()
        else:
            raise NotImplementedError





from cuqi.implicitprior import RegularizedGaussian
from cuqi.distribution import Distribution, GMRF

class RegularizedGMRF(RegularizedGaussian):
    """ Implicit Regularized GMRF (Gaussian Markov Random Field). 

    Defines a so-called implicit prior based on a GMRF distribution with implicit regularization.
    The regularization can be defined in the form of a proximal operator or a projector.
    Alternatively, preset constraints and regularization can be used.

    Only one of proximal, projector, constraint or regularization can be provided. Otherwise, an error is raised.

    Can be used as a prior in a posterior which can be sampled with the RegularizedLinearRTO sampler.


    For more details on implicit regularized Gaussian see the following paper:

    [1] Everink, Jasper M., Yiqiu Dong, and Martin S. Andersen. "Sparse Bayesian inference with regularized
    Gaussian distributions." Inverse Problems 39.11 (2023): 115004.

    Parameters
    ----------
    mean
        See :class:`~cuqi.distribution.GMRF` for details.
        
    prec
        See :class:`~cuqi.distribution.GMRF` for details.

    bc_type
        See :class:`~cuqi.distribution.GMRF` for details.

    order
        See :class:`~cuqi.distribution.GMRF` for details.

    proximal : callable f(x, scale) or None
        Euclidean proximal operator f of the regularization function g, that is, a solver for the optimization problem
        min_z 0.5||x-z||_2^2+scale*g(x).

    projector : callable f(x) or None
        Euclidean projection onto the constraint C, that is, a solver for the optimization problem
        min_(z in C) 0.5||x-z||_2^2.

    constraint : string or None
        Preset constraints. Can be set to "nonnegativity" and "box". Required for use in Gibbs.
        For "box", the following additional parameters can be passed:
            lower_bound : array_like or None
                Lower bound of box, defaults to zero
            upper_bound : array_like
                Upper bound of box, defaults to one

    regularization : string or None
        Preset regularization. Can be set to "l1". Required for use in Gibbs in future update.
        For "l1", the following additional parameters can be passed:
            strength : scalar
                Regularization parameter, i.e., strength*||x||_1 , defaults to one

    """
    def __init__(self, mean=None, prec=None, bc_type='zero', order=1, proximal = None, projector = None, constraint = None, regularization = None, **kwargs):
            
            args = {"lower_bound" : kwargs.pop("lower_bound", None),
                    "upper_bound" : kwargs.pop("upper_bound", None),
                    "strength" : kwargs.pop("strength", None)}
            
            # Underlying explicit GMRF
            self._gaussian = GMRF(mean, prec, bc_type=bc_type, order=order, **kwargs)
            
            # Init from abstract distribution class
            super(Distribution, self).__init__(**kwargs)

            self._parse_regularization_input_arguments(proximal, projector, constraint, regularization, args)



import numpy as np
import cuqi
import pytest
import matplotlib
import matplotlib.pyplot as plt
from cuqi import geometry
from cuqi.samples import Samples

@pytest.mark.parametrize("to_funvals, plot_par",
    [(True,False),(True,True),(False,False),(False,True)])
@pytest.mark.parametrize("geom",[
						(cuqi.geometry.Discrete(1)),
						(cuqi.geometry.Discrete(3)),
						(cuqi.geometry.Continuous1D(1)),
						(cuqi.geometry.Continuous1D(3)),
                        (cuqi.geometry.Image2D((3,3))),
						])
def test_samples_plot(geom, to_funvals, plot_par):
    # Make basic distribution and sample
    dim = geom.par_dim
    x = cuqi.distribution.Normal(np.zeros(dim),np.ones(dim),geometry=geom)
    s = x.sample(10)

    # Convert to funvals if requested
    if to_funvals:
        s = s.funvals

    # Verify that asking for plotting in the parameter space when samples
    # are funvals raises error
    if to_funvals and plot_par:
        with pytest.raises(Exception, match=r"Plot par is true, but is_par is false"):
            s.plot(plot_par=plot_par)
        return
    
    #Verify plotting of single and multiple samples
    s.plot(plot_par=plot_par)
    s.plot(0,plot_par=plot_par)

    # Raise error when calling plot_chain if the samples are not in vector form
    if not s.is_vec:
        with pytest.raises(Exception, match=r"Cannot perform plot_chain"):
            s.plot_chain()
        # Convert the samples to vector form
        s = s.vector
    
    #s.plot_chain() #No default selection of chain
    s.plot_chain(0)
    if dim > 2:
        s.plot([0,2],plot_par=plot_par)
        s.plot_chain([0,2])
    
    # Verify passing is_par raises error (because is_par is determined by
    # the type of the samples)
    with pytest.raises(ValueError,match=r"Cannot pass is_par as a plotting argument"):
        s.plot(is_par=not to_funvals,plot_par=plot_par)

@pytest.mark.parametrize("to_funvals, plot_par", [(True,False),(True,True),(False,False),(False,True)])
@pytest.mark.parametrize("plot_method, plot_par_supported, arviz",
                         [(Samples.plot_autocorrelation, False, True),
                          (Samples.plot_trace, False, True),
                          (Samples.plot_violin, False, True),
                          (Samples.plot_pair, False, True),
                          (Samples.plot_std, True, False),
                          (Samples.plot_mean, True, False),
                          (Samples.plot_median, True, False),
                          (Samples.plot_variance, True, False),
                          (Samples.plot_ci_width, True, False)])
@pytest.mark.parametrize("geom",[
						(cuqi.geometry.Discrete(3)),
						(cuqi.geometry.Continuous1D(4)),
                        (cuqi.geometry.Image2D((4,4))),
                        (cuqi.geometry.MappedGeometry(
                             cuqi.geometry.Image2D((3,3)),
                             map=lambda x: x**2+2))
						])
def test_samples_plot_methods(geom, to_funvals, plot_par,
                              plot_method, plot_par_supported,
                              arviz):
    """Test Samples class plotting methods"""

    # Make basic distribution and sample
    np.random.seed(0)
    dim = geom.par_dim
    x = cuqi.distribution.Normal(np.zeros(dim),np.ones(dim),geometry=geom)
    s = x.sample(10)

    # Convert to funvals if requested
    if to_funvals:
        s = s.funvals

    # Verify that using an arviz based method when the samples are not in
    # vector form raises error
    if arviz and not s.is_vec:
        with pytest.raises(Exception, match=r"Samples are not in a vector representation"):
            plot_method(s)
        # Convert the samples to vector form
        s = s.vector
  
    # Verify asking for plot_par when samples are funvals raises error
    if to_funvals and plot_par:
        if plot_par_supported:
            # The case of plotting methods that support plot_par
            with pytest.raises(Exception, match=r"Plot par is true, but is_par is false"):
                plot_method(s, plot_par=plot_par)
            return
        else:
            # The case of plotting methods that do not support plot_par
            with pytest.raises(TypeError, match=r"got an unexpected keyword argument"):
                plot_method(s, plot_par=plot_par)
            return
    
    # Verify passing is_par raises error
    if plot_par_supported:
        with pytest.raises(ValueError,match=r"Cannot pass is_par as a plotting argument"):
            plot_method(s, is_par=not to_funvals)
    else:
        with pytest.raises(TypeError,match=r"got an unexpected keyword argument"):
            plot_method(s, is_par=not to_funvals)
    
    #Verify plotting method works
    if not plot_par_supported:
        plot_method(s)
    else:
        plot_method(s, plot_par=plot_par)


@pytest.mark.parametrize("kwargs",[
                        ({}),
                        ({"max_lag":10,"textsize":25}),
                        ])
def test_samples_plot_autocorrelation(kwargs):
    # Make basic distribution and sample
    dist = cuqi.distribution.DistributionGallery("CalSom91")
    sampler = cuqi.sampler.MH(dist)
    samples = sampler.sample_adapt(1000)

    # Switch to discrete geometry (easiest for "variable" names)
    samples.geometry = cuqi.geometry.Discrete(["alpha","beta"])

    # Plot with defaults
    samples.plot_autocorrelation()

    # Plot with defaults arguments
    samples.plot_autocorrelation(**kwargs)

    # Plot for single parameter + arguments
    samples.plot_autocorrelation([0],**kwargs)


@pytest.mark.parametrize("kwargs",[
                        ({}),
                        ({"combined":False,"tight_layout":False}),
                        ({"exact":[1, 1]}) # Check exact can be compared with
                        ])
def test_samples_plot_trace(kwargs):
    # Make basic distribution and sample
    dist = cuqi.distribution.DistributionGallery("CalSom91")
    sampler = cuqi.sampler.MH(dist)
    samples = sampler.sample_adapt(1000)

    # Switch to discrete geometry (easiest for "variable" names)
    samples.geometry = cuqi.geometry.Discrete(["alpha","beta"])

    # Plot with defaults
    samples.plot_trace()

    # Plot with defaults arguments
    samples.plot_trace(**kwargs)

    # Plot for single parameter + arguments
    samples.plot_trace([0],**kwargs)

@pytest.mark.parametrize("kwargs",[
                        ({"marginals": False}),
                        ({"marginals": True}),
                        ({"kind":"kde","marginals":False}),
                        ({"kind":"kde","marginals":True}),
                        ({"kind":"hexbin","marginals":False}),
                        ({"kind":"hexbin","marginals":True}),
                        ])
def test_samples_plot_pair(kwargs):
    # Make basic distribution and sample
    dist = cuqi.distribution.Gaussian(np.array([1,2,3,4]),1)
    sampler = cuqi.sampler.MH(dist)
    samples = sampler.sample_adapt(1000)
    samples.geometry = cuqi.geometry.Discrete(["a","b","c","d"])

    # Plot with defaults
    samples.plot_pair()

    # Plot with defaults arguments
    samples.plot_pair(**kwargs)

    # Plot for single parameter + arguments
    ax = samples.plot_pair([1,2],**kwargs)

    # Check if the correct labels are passed in specific case
    if kwargs.get("marginals") == False:
        assert ax.get_xlabel() == "b"
        assert ax.get_ylabel() == "c"

def test_rhat_values():
    rng = np.random.RandomState(0)
    mean = 0; var  = 1
    samples1 = cuqi.distribution.Normal(mean,var).sample(20000,rng=rng)
    samples2 = cuqi.distribution.Normal(mean,var).sample(20000,rng=rng) 
    rhat_results = samples1.compute_rhat(samples2)
    assert np.allclose(rhat_results[0], 1, rtol=1e-3)

def test_rhat_geometry():
    mean = 0; var  = 1
    samples1 = cuqi.distribution.Normal(mean,var).sample(200)
    samples2 = cuqi.distribution.Normal(mean,var).sample(200) 
    samples1.geometry = cuqi.geometry.Discrete(["alpha","beta1"])
    samples2.geometry = cuqi.geometry.Discrete(["alpha","beta2"])
    with pytest.raises(TypeError): #Type error since geometry does not match.
        samples1.compute_rhat(samples2)

def test_ess():
    dist = cuqi.distribution.DistributionGallery("CalSom91")
    sampler = cuqi.sampler.MH(dist)
    samples = sampler.sample_adapt(500)
    assert samples.compute_ess().shape == samples.geometry.par_shape

@pytest.mark.parametrize("geometry", [cuqi.geometry.Discrete(2),
                                      cuqi.geometry.MappedGeometry(
                                          cuqi.geometry.Continuous1D(2), map=lambda x: x**2),
                                      cuqi.geometry.KLExpansion(np.arange(0, 1, .1), num_modes=4),
                                      cuqi.geometry.Image2D((3,3)),
                                      cuqi.geometry.MappedGeometry(
                                          cuqi.geometry.Image2D((4,3)), map=lambda x: x**2+1)])
def test_samples_funvals(geometry):
    """Test that the function values are computed correctly."""
    # Generate some random samples
    Ns = 10
    samples = cuqi.samples.Samples(
        np.random.randn(geometry.par_dim, Ns), geometry=geometry)

    # Compute the function values
    funvals_vec = np.empty((geometry.funvec_dim, Ns))
    for i, s in enumerate(samples):
        funvals_vec[:, i] = geometry.fun2vec(geometry.par2fun(s))

    # Check that the funvals property is implemented correctly
    assert np.allclose(samples.funvals.vector.samples, funvals_vec)


@pytest.mark.parametrize("percent", [10, 50, 90, 95, 99])
@pytest.mark.parametrize("compute_on_par", [False, True])
@pytest.mark.parametrize("geometry", [cuqi.geometry.Discrete(2),
                                      cuqi.geometry.MappedGeometry(
                                        cuqi.geometry.Continuous1D(2), map=lambda x: x**2),
                                      cuqi.geometry.Image2D((2,1)),
                                      cuqi.geometry.MappedGeometry(
                                          cuqi.geometry.Image2D((1,2)), map=lambda x: x**2+1)])
def test_compute_ci(percent, compute_on_par, geometry):
    dist = cuqi.distribution.DistributionGallery("CalSom91")
    sampler = cuqi.sampler.MH(dist)
    par_samples = sampler.sample_adapt(500)
    par_samples.geometry = geometry
    
    samples = par_samples if compute_on_par else par_samples.funvals
    ci = samples.compute_ci(percent)

    # manually compute ci
    lb = (100-percent)/2
    up = 100-lb
    lo_conf, up_conf = np.percentile(samples.samples, [lb, up], axis=-1)

    assert np.allclose(ci[0], lo_conf)
    assert np.allclose(ci[1], up_conf)

@pytest.mark.parametrize("is_par", [False, True, None]) #passing is_par will raise an error.
@pytest.mark.parametrize("plot_par, compute_on_par",
                         [(True, True),
                          (True, False), # This case will raise an error.
                          (False, True),
                          (False, False)])
@pytest.mark.parametrize("geometry", [cuqi.geometry.Discrete(2),
                                      cuqi.geometry.KLExpansion(np.arange(0, 1, .1))])
def test_plot_ci_par_func(is_par, plot_par, compute_on_par, geometry):
    """Test passing flags to plot_ci."""
    np.random.seed(0)
    par_samples = cuqi.samples.Samples(np.random.randn(geometry.par_dim, 10), geometry=geometry)
    samples = par_samples if compute_on_par else par_samples.funvals

    if is_par is not None:
        # User should not be able to pass is_par for plotting ci because it is
        #  determined automatically depending on self.is_par and plot_par
        with pytest.raises(ValueError):
            samples.plot_ci(is_par=is_par, plot_par=plot_par)

    elif plot_par and not compute_on_par:
        # User cannot ask for computing statistics on function values then plotting on parameter space
        # plot_ci will raise an error in this case
        with pytest.raises(ValueError):
            samples.plot_ci(plot_par=plot_par)
    else:
        #The remaining cases should not raise an error.
        import matplotlib.pyplot as plt
        plt.figure()
        samples.plot_ci(plot_par=plot_par)


@pytest.mark.parametrize("geometry",
                         [cuqi.geometry.Discrete(4),
                          cuqi.geometry.KLExpansion(np.linspace(0, 1, 4)),
                          cuqi.geometry.Continuous2D((2, 2)),
                          cuqi.geometry.Image2D((2, 2))])
@pytest.mark.parametrize("plot_par", [False, True])
def test_plot_ci_returned_values(geometry, plot_par):
    """Test that the correct matplotlib object types are returned by plot_ci."""
    # Create samples and plot ci
    np.random.seed(0)
    samples = cuqi.samples.Samples(np.random.randn(
        geometry.par_dim, 10), geometry=geometry)
    plot_objs = samples.plot_ci(
        100, plot_par=plot_par, exact=np.random.randn(geometry.par_dim))

    # Check that the correct matplotlib object types are returned:
    # Continuous 2D plots
    if isinstance(geometry, cuqi.geometry.Continuous2D) and not plot_par: 
        for obj in plot_objs:
            assert isinstance(obj, matplotlib.collections.PolyCollection)

    # Image 2D plots
    elif isinstance(geometry, cuqi.geometry.Image2D) and not plot_par: 
        for obj in plot_objs:
            assert isinstance(obj, matplotlib.image.AxesImage)

    # Discrete plots
    elif plot_par or isinstance(geometry, cuqi.geometry.Discrete):
        assert isinstance(plot_objs[0], matplotlib.lines.Line2D)
        assert isinstance(plot_objs[1], matplotlib.lines.Line2D)
        assert isinstance(plot_objs[2], matplotlib.container.ErrorbarContainer)

    # Continuous 1D plots
    elif plot_par or isinstance(geometry, cuqi.geometry.Discrete):  
        assert isinstance(plot_objs[0], matplotlib.lines.Line2D)
        assert isinstance(plot_objs[1], matplotlib.lines.Line2D)
        assert isinstance(plot_objs[2], matplotlib.collections.PolyCollection)


@pytest.mark.parametrize("geom, map, imap, supported",
                         [(cuqi.geometry.Discrete(4), lambda x:x**2, lambda x:np.sqrt(x), True),
                          (cuqi.geometry.Continuous1D(15),
                           lambda x:x+12, lambda x:x-12, True),
                             (cuqi.geometry.Image2D((4, 5)), lambda x:x **
                              2+1, lambda x:np.sqrt(x-1), True)
                          ])
def test_parameters_property(geom, map, imap, supported):
    """Test that the Samples parameters property is computed correctly. And that an error is generated
    when the Samples geometry type does not support computing funvals."""
    # Create random samples:
    np.random.seed(0)
    Ns = 10
    mapped_geom = cuqi.geometry.MappedGeometry(geom, map, imap)
    val = np.absolute(np.random.rand(mapped_geom.par_dim, Ns))

    # Create Samples object:
    samples = cuqi.samples.Samples(val, geometry=mapped_geom)

    if not supported:
        with pytest.raises(ValueError, match=r"Cannot convert the samples to function values"):
            funvals = samples.funvals
    else:
        # Compute function values and from the function values
        # compute the parameters:
        funvals_vec = samples.funvals.vector
        parameters = funvals_vec.funvals.parameters

        # Assert that the parameters and the function values (in vector 
        # representation) are different and that extracting the function values
        # and going back to the parameters is done correctly.
        assert not np.allclose(parameters.samples, funvals_vec.samples) and\
            np.allclose(parameters.samples, val)

def test_cuqiarray_default_geometry():
    """ Test that CUQIarray creates a default geometry when no geometry is passed"""
    v = cuqi.array.CUQIarray([0,1,2,3,4,5,6,7,8])
    assert type(v.geometry) is cuqi.geometry._geometry._DefaultGeometry1D

def test_cuqiarray_multidim():
    X = np.array([[1, 2], [3,4]])
    with pytest.raises(Exception) as e:
        C = cuqi.array.CUQIarray(X)
    assert "input_array cannot be multidimensional when initializing CUQIarray as parameter (with is_par True)." in str(e.value) # this message
    assert e.type == ValueError  

def test_cuqiarray_ispar_false_without_geometry():
    with pytest.raises(Exception) as e:
        X = cuqi.array.CUQIarray([1,2,3], is_par=False)
    assert "geometry cannot be none when initializing a CUQIarray as function values (with is_par False)." in str(e.value) # this message
    assert e.type == ValueError   

def test_violin_plot():
    """ Test that the violin plot is generated correctly. """
    dist = cuqi.distribution.DistributionGallery("CalSom91")
    sampler = cuqi.sampler.MH(dist)
    samples = sampler.sample_adapt(1000)
    samples.geometry = cuqi.geometry.Discrete(["alpha","beta"])

    ax = samples.plot_violin(shade=0.1)

    assert ax[0].get_title() == "alpha"
    assert ax[1].get_title() == "beta"

@pytest.mark.parametrize("method", [Samples.mean,
                                    Samples.variance,
                                    Samples.std,
                                    Samples.std,
                                    Samples.ci_width])
def test_computing_statistics_on_arbitrary_objects_raises_error(method):
    """ Test that computing statistics on arbitrary objects (np.matrix object for example) raises an error. """

    samples = cuqi.samples.Samples([object(), object()])
    with pytest.raises(TypeError, match=r"Cannot compute statistics"):
        method(samples)



from cuqi.distribution import Posterior, LMRF, Gamma
import numpy as np
import scipy as sp

class ConjugateApprox: # TODO: Subclass from Sampler once updated
    """ Approximate Conjugate sampler

    Sampler for sampling a posterior distribution where the likelihood and prior can be approximated
    by a conjugate pair.

    Currently supported pairs are:
    - (LMRF, Gamma): Approximated by (Gaussian, Gamma)

    For more information on conjugate pairs, see https://en.wikipedia.org/wiki/Conjugate_prior.

    """

    
    def __init__(self, target: Posterior):
        if not isinstance(target.likelihood.distribution, LMRF):
            raise ValueError("Conjugate sampler only works with Laplace diff likelihood function")
        if not isinstance(target.prior, Gamma):
            raise ValueError("Conjugate sampler only works with Gamma prior")
        self.target = target

    def step(self, x=None):
        # Extract variables
        # Here we approximate the Laplace diff with a Gaussian

        # Extract diff_op from target likelihood
        D = self.target.likelihood.distribution._diff_op
        n = D.shape[0]

        # Gaussian approximation of LMRF prior as function of x_k
        # See Uribe et al. (2022) for details
        # Current has a zero mean assumption on likelihood! TODO
        beta=1e-5
        def Lk_fun(x_k):
            dd =  1/np.sqrt((D @ x_k)**2 + beta*np.ones(n))
            W = sp.sparse.diags(dd)
            return W.sqrt() @ D

        x = self.target.likelihood.data                 #x
        d = len(x)                                      #d
        Lx = Lk_fun(x)@x                                #Lx
        alpha = self.target.prior.shape                 #alpha
        beta = self.target.prior.rate                   #beta

        # Create Gamma distribution and sample
        dist = Gamma(shape=d+alpha, rate=np.linalg.norm(Lx)**2+beta)

        return dist.sample()



import numpy as np
from scipy.linalg import toeplitz
from scipy.sparse import csc_matrix
from scipy.integrate import quad_vec
from scipy.signal import fftconvolve
from scipy.ndimage import convolve1d

import cuqi
from cuqi.model import LinearModel
from cuqi.distribution import Gaussian
from cuqi.problem import BayesianProblem
from cuqi.geometry import Geometry, MappedGeometry, StepExpansion, KLExpansion, KLExpansion_Full, CustomKL, Continuous1D, Continuous2D, Image2D
from cuqi.array import CUQIarray
import warnings



#=============================================================================
class _Deblur(BayesianProblem):
    """
    1D Deblur test problem.

    Parameters
    ------------
    dim : int, default 128
        size of the (dim,dim) deblur problem.

    bounds : len=2 list of int's, default [0,1]
        Lower and upper bounds for the mesh.

    blur_size : int, default 48
        size of blur.

    noise_std : scalar, default 0.1
        Standard deviation of the noise.

    prior : cuqi.distribution.Distribution, default Gaussian
        Distribution of the prior.

    Attributes
    ----------
    data : ndarray
        Generated (noisy) data

    model : cuqi.model.Model
        Deblur forward model

    likelihood : cuqi.likelihood.Likelihood
        Likelihood function.

    prior : cuqi.distribution.Distribution, Default Gaussian
        Distribution of the prior

    exactSolution : ndarray
        Exact solution (ground truth)

    exactData : ndarray
        Noise free data

    mesh : ndarray
        The mesh the model is defined on.

    meshsize : float
        Size of each mesh element.

    Methods
    ----------
    MAP()
        Compute MAP estimate of posterior.
        NB: Requires prior to be defined.

    sample_posterior(Ns)
        Sample Ns samples of the posterior.
        NB: Requires prior to be defined.

    """
    def __init__(self, dim=128, bounds=[0, 1], blur_size=48, noise_std=0.1, prior=None):
        
        warnings.warn("DEPRECATED: Use Deconvolution1D instead.")
        
        # mesh
        mesh = np.linspace(bounds[0], bounds[1], dim)
        meshsize = mesh[1] - mesh[0]

        # set-up computational model kernel
        kernel = lambda x, y, blur_size: blur_size / 2*np.exp(-blur_size*abs((x-y)))   # blurring kernel

        # convolution matrix
        T1, T2 = np.meshgrid(mesh, mesh)
        A = meshsize*kernel(T1, T2, blur_size)
        maxval = A.max()
        A[A < 5e-3*maxval] = 0
        A = csc_matrix(A)   # make A sparse

        # Store forward model
        model = LinearModel(A)
        
        # Prior
        if prior is None:
            prior = Gaussian(np.zeros(dim), 1, name="x")

        # Store data distribution
        data_dist = Gaussian(model(prior), noise_std**2, name="y")
        
        # Generate inverse-crime free data (still same blur size)
        data, f_true, g_true = self._generateData(mesh, kernel, blur_size, data_dist)

        # Likelihood
        likelihood = data_dist.to_likelihood(data)

        #Initialize deblur as BayesianProblem cuqi problem
        super().__init__(likelihood, prior)
        
        #Store other properties
        self.meshsize = meshsize
        self.exactSolution = f_true
        self.exactData = g_true
        self.mesh = mesh
        
    def _generateData(self,mesh,kernel,blur_size,data_dist):

        # f is piecewise constant
        x_min, x_max = mesh[0], mesh[-1]
        vals = np.array([0, 2, 3, 2, 0, 1, 0])
        conds = lambda x: [(x_min <= x) & (x < 0.1), (0.1 <= x) & (x < 0.15), (0.15 <= x) & (x < 0.2),  \
                (0.20  <= x) & (x < 0.25), (0.25 <= x) & (x < 0.3), (0.3 <= x) & (x < 0.6), \
                (0.6 <= x) & (x <= x_max)]
        f_signal = lambda x: np.piecewise(x, conds(x), vals)

        # numerically integrate the convolution
        g_conv = lambda x: quad_vec(lambda y: f_signal(y)*kernel(x, y, blur_size), x_min, x_max)
        # se also np.convolve(kernel(...), f_true, mode='same')

        # true values
        f_true = f_signal(mesh)
        g_true = g_conv(mesh)[0]

        # noisy data
        data = g_true + data_dist(np.zeros(len(mesh))).sample() #np.squeeze(noise.sample(1)) #np.random.normal(loc=0, scale=self.sigma_obs, size=(self.dim))

        return data, f_true, g_true

#=============================================================================
class Deconvolution1D(BayesianProblem):
    """
    
    Create a 1D periodic deconvolution test problem defined by the inverse problem

    .. math::

        \mathbf{b} = \mathbf{A}\mathbf{x},

    where :math:`\mathbf{b}` is a (noisy) convolved signal,
    :math:`\mathbf{x}` is a sharp (clean) signal and
    :math:`\mathbf{A}` is a convolution operator.

    The convolution operator is defined by specifying a point spread function and
    boundary conditions and is computed via scipy.ndimage.convolve1D. By default,
    the matrix representation of the convolution operator is computed and stored.

    The inputs are padded to fit the boundary conditions.
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.ndimage.convolve1d.html

    Parameters
    ------------
    dim : int, default 128
        size of the (dim,dim) deconvolution problem

    PSF : string or ndarray, default 'Gauss'
        | Determines type of the underlying point spread function (PSF).
        | Depending if use_legacy is True or False, the following options are available:
        | 'Gauss' - a 1D Gaussian blur function
        | 'Moffat' - a 1D Moffat blur function (non-LEGACY only)
        | 'Defocus' - an out-of-focus 1D blur function (non-LEGACY only)
        | 'sinc' or 'prolate' - a sinc function (LEGACY only)
        | 'vonMises' - a periodic version of the Gauss function (LEGACY only)
        | ndarray - a custom PSF represented as a 1D ndarray.

    PSF_param : scalar, default depends on PSF
        | A parameter that determines the shape of the PSF;
        | the larger the parameter, the larger the blur on the signal.
        | Ignored if PSF is a ndarray.

    PSF_size : int, default equal to dim
        | The size of the PSF.
        | Ignored if PSF is a ndarray.

    BC : string, default 'periodic'
        | Boundary conditions for the convolution.
        | 'zero' - zero boundary conditions
        | 'periodic' - periodic boundary conditions
        | 'Mirror' - Reflected around center of last pixel boundary conditions
        | 'Reflect' - Reflected around edge of last pixel boundary conditions
        | 'Nearest' - Replicates last element of boundary

    phantom : string or ndarray, default 'sinc'
        | The phantom that is sampled to produce the exact solution (signal).
        | 'Gauss' - a Gaussian function
        | 'sinc' - a sinc function
        | 'vonMises' - a periodic version of the Gauss function
        | 'square' - a "top hat" function
        | 'hat' - a triangular hat function
        | 'bumps' - two bumps
        | 'derivGauss' - the first derivative of Gauss function
        | 'pc' - Piece-wise constant phantom
        | 'skyscraper' - Piece-wise constant phantom with multiple peaks
        | ndarray - a custom phantom

    phantom_param : scalar, default depends on phantom
        | A parameter that determines the horizontal scaling of the
        | function; the larger the parameter the more horizontally
        | compressed. Does not apply to phantom = 'bumps', 'pc', 'skyscraper' or ndarray.

    noise_type : string, default 'gaussian'
        | The type of noise
        | "Gaussian" - Additive Gaussian white noise¨
        | "scaledGaussian" - Gaussian noise with standard deviation
        |                    scaled by magnitude of the data for each
        |                    point.

    noise_std : scalar, default 0.01
        Standard deviation of the noise

    prior : cuqi.distribution.Distribution, Default Gaussian
        Distribution of the prior

    use_legacy : bool, Default False
        If True, use the legacy matrix representation of the forward model.
        The legacy representation has different choices for the PSF.
        The data is scaled differently than the non-legacy representation.
        

    """
    def __init__(self,
        dim=128,
        PSF="gauss",
        PSF_param=None,
        PSF_size=None,
        BC="periodic",
        phantom="sinc",
        phantom_param=None,
        noise_type="gaussian",
        noise_std=0.01,
        prior=None,
        use_legacy=False,
        ):
        
        # Set up forward model
        if use_legacy: # Legacy matrix representation

            if BC != "periodic":
                raise ValueError("Legacy matrix representation only supports periodic boundary conditions")

            if PSF_size is not None:
                raise ValueError("Legacy matrix representation does not support PSF_size")

            A = _getCirculantMatrix(dim, PSF, PSF_param)

            model = cuqi.model.LinearModel(A, range_geometry=Continuous1D(dim), domain_geometry=Continuous1D(dim))

        else: # New convolution operator based on scipy.ndimage.convolve1d

            Afun = _getConvolutionOperator(dim, PSF, PSF_param, PSF_size, BC)

            # For efficiency, we create a sparse matrix representation of the forward model,
            # instead of using the matrix-free representation. For 1D problems, there is
            # no need to use the matrix-free representation, since the matrix is small.
            Id = np.eye(dim)
            A = np.array([Afun(Id[:, i]) for i in range(dim)])
            A = csc_matrix(A) # make it sparse

            model = cuqi.model.LinearModel(A, range_geometry=Continuous1D(dim), domain_geometry=Continuous1D(dim))  

        # Set up exact solution
        if isinstance(phantom, np.ndarray):
            if phantom.ndim != 1 or phantom.shape[0] != dim:
                raise ValueError("phantom must be a 1D array of length dim")
            if phantom_param is not None:
                warnings.warn("phantom_param is ignored when phantom is a ndarray")
            x_exact = phantom
        elif isinstance(phantom, str):
            x_exact = _getExactSolution(dim, phantom, phantom_param)
        else:
            raise ValueError(f"Unknown phantom type {phantom}")

        x_exact = CUQIarray(x_exact, geometry=model.domain_geometry)

        # Generate exact data
        y_exact = model.forward(x_exact)

        # Set up prior
        if prior is None:
            prior = cuqi.distribution.Gaussian(np.zeros(dim), 1, name="x")

        # Define and add noise #TODO: Add Poisson and logpoisson
        if noise_type.lower() == "gaussian":
            data_dist = cuqi.distribution.Gaussian(model(prior), noise_std**2, name="y")
        elif noise_type.lower() == "scaledgaussian":
            data_dist = cuqi.distribution.Gaussian(model(prior), (y_exact*noise_std)**2, name="y")
        else:
            raise NotImplementedError("This noise type is not implemented")
        
        # Generate data
        data = data_dist(x_exact).sample()

        # Likelihood
        likelihood = data_dist.to_likelihood(data)

        # Set up as Bayesian problem
        super().__init__(likelihood, prior)

        # Store exact values
        self.exactSolution = x_exact
        self.exactData = y_exact
        self.infoString = "Noise type: Additive {} with std: {}".format(noise_type.capitalize(),noise_std)

def _getConvolutionOperator(dim, PSF, PSF_param, PSF_size, BC):

    # Boundary condition translation
    if BC.lower() == "zero":
        mode = "constant"
    elif BC.lower() == "periodic":
        mode = "wrap"
    elif BC.lower() == "mirror":
        mode = "mirror"
    elif BC.lower() == "reflect":
        mode = "reflect"
    elif BC.lower() == "nearest":
        mode = "nearest"
    else:
        raise ValueError("Unknown boundary condition")

    if PSF_size is None:
        PSF_size = dim
   
    # PSF setup
    if isinstance(PSF, np.ndarray):
        if PSF.ndim != 1:
            raise ValueError("PSF must be a 1D array")
        P = PSF
    elif isinstance(PSF, str):
        if PSF.lower() == "gauss":
            P, _ = _GaussPSF_1D(PSF_size, PSF_param)
        elif PSF.lower() == "moffat":
            P, _ = _MoffatPSF_1D(PSF_size, PSF_param)
        elif PSF.lower() == "defocus":
            P, _ = _DefocusPSF_1D(PSF_size, PSF_param)
        else:
            raise ValueError("Unknown PSF type")
    
    # Convolution matrix
    return lambda x: convolve1d(x, P, mode=mode)


def _createPSF_1D(PSF_size, PSF_func):
    """ Create a 1D normalized PSF of size PSF_size using the function PSF_func. """
    # Set up grid points
    x = np.arange(-np.fix(PSF_size/2), np.ceil(PSF_size/2))

    # Compute the PSF
    PSF = PSF_func(x)

    # Normalize the PSF.
    PSF /= PSF.sum()

    # find the center
    center = np.where(PSF == PSF.max())[0][0]
    return PSF, center.astype(int)

def _GaussPSF_1D(PSF_size, PSF_param):
    """ Create a 1D normalized Gaussian PSF of size PSF_size with standard deviation PSF_param. """

    # Set default value for PSF_param
    if PSF_param is None:
        PSF_param = 10

    # Set up Gaussian function
    PSF_func = lambda x: np.exp( -0.5*((x**2)/(PSF_param**2)) )

    return _createPSF_1D(PSF_size, PSF_func)

def _MoffatPSF_1D(PSF_size, PSF_param, beta=1):
    """ Create a 1D normalized Moffat PSF of size PSF_size with standard deviation PSF_param. """

    # Set default value for PSF_param
    if PSF_param is None:
        PSF_param = 10

    # Set up Moffat function
    PSF_func = lambda x: ( 1 + (x**2)/(PSF_param**2) )**(-beta)

    return _createPSF_1D(PSF_size, PSF_func)

def _DefocusPSF_1D(PSF_size, PSF_param):
    """ Create a 1D normalized defocus PSF of size PSF_size with standard deviation PSF_param. """

    # Set default value for PSF_param
    if PSF_param is None:
        PSF_param = 10

    center = np.fix(int(PSF_size/2))
    if (PSF_param == 0):    
        # the PSF is a delta function and so the blurring matrix is I
        PSF = np.zeros(PSF_size)
        PSF[center] = 1
    else:
        PSF = np.ones(PSF_size) / (np.pi * PSF_param**2)
        k = np.arange(1, PSF_size+1)
        aa = (k-center)**2
        idx = np.array((aa > (PSF_param**2)))
        PSF[idx] = 0
    PSF = PSF / PSF.sum()

    return PSF, center.astype(int)

def _getCirculantMatrix(dim, PSF, PSF_param):
    """
    GetCircMatrix  Create a circulant matrix for deconvolution examples
    
    _getCirculantMatrix(dim, PSF, PSF_param)
    
    Input:  dim = size of the circulant matrix A
            PSF = string that determined type of the underlying PSF
                    'Gauss' - a Gaussian function
                    'sinc' or 'prolate' - a sinc function
                    'vonMises' - a periodic version of the Gauss function
                    ndarray - a custom PSF.
            PSF_param = a parameter that determines the shape of the PSF;
                    the larger the parameter, to slower the initial decay
                    of the singular values of A

    Output: circulant matrix

    Based on Matlab code by Per Christian Hansen, DTU Compute, April 7, 2021
    """

    if not (dim % 2) == 0:
        raise NotImplementedError("Circulant matrix not implemented for odd numbers")

    if isinstance(PSF, np.ndarray):
        if PSF_param is not None: warnings.warn("PSF_param is ignored when PSF is a ndarray")
        if PSF.ndim != 1 or PSF.shape[0] != dim:
            raise ValueError("kernel must be a 1D array of length dim")
        h = np.roll(PSF, -int(dim/2))
        #h = h/np.linalg.norm(h)**2 # TODO: Normalize
        hflip = np.concatenate((h[0:1], np.flipud(h[1:])))
        return toeplitz(hflip,h) 

    dim_half = dim/2
    grid = np.arange(dim_half+1)/dim

    if PSF.lower() == "gauss":
        if PSF_param is None: PSF_param = 10
        h = np.exp(-(PSF_param*grid)**2)
        h = np.concatenate((h,np.flipud(h[1:-1])))
        hflip = np.concatenate((h[0:1], np.flipud(h[1:])))
        return toeplitz(hflip,h)    

    elif PSF.lower() == "sinc" or PSF.lower() == "prolate":
        if PSF_param is None: PSF_param = 15
        h = np.sinc(PSF_param*grid)
        h = np.concatenate((h,np.flipud(h[1:-1])))
        hflip = np.concatenate((h[0:1], np.flipud(h[1:])))
        return toeplitz(hflip,h)

    elif PSF.lower() == "vonmises":
        if PSF_param is None: PSF_param = 5
        h = np.exp(np.cos(2*np.pi*grid))
        h = (h/h[0])**PSF_param
        h = np.concatenate((h,np.flipud(h[1:-1])))
        hflip = np.concatenate((h[0:1], np.flipud(h[1:])))
        return toeplitz(hflip,h)

    else:
        raise NotImplementedError(f"PSF {PSF.lower()} is not implemented for the legacy convolution operator.")

def _getExactSolution(dim, phantom, phantom_param):
    """
    GetExactSolution  Create a periodic solution
    
     Input: dim = length of the vector x
            phantom = the phantom that is sampled to produce x
                    'Gauss' - a Gaussian function
                    'sinc' - a sinc function
                    'vonMises' - a periodic version of the Gauss function
                    'square' - a "top hat" function
                    'hat' - a triangular hat function
                    'bumps' - two bumps
                    'derivGauss' - the first derivative of Gauss function
                    'pc' - Piece-wise constant phantom with one tall peak and one wide peak
                    'skyscraper' - Piece-wise constant phantom with multiple peaks of varying height
            param = A parameter that determines the horizontal scaling of the
                    function; the larger the parameter the more horizontally
                    compressed. Does not apply to phantom = 'bumps' or ndarray.
    
    Output: x = column vector, scaled such that max(x) = 1

    Based on Matlab code by Per Christian Hansen, DTU Compute, April 7, 2021
    """

    if phantom.lower() == "gauss":
        if phantom_param is None: phantom_param = 5
        return np.exp(-(phantom_param*np.linspace(-1,1,dim))**2)

    elif phantom.lower() == "sinc":
        if phantom_param is None: phantom_param = 5
        return np.sinc(phantom_param*np.linspace(-1,1,dim))        

    elif phantom.lower() == "vonmises":
        if phantom_param is None: phantom_param = 5
        x = np.exp(np.cos(np.pi*np.linspace(-1,1,dim)))
        return (x/np.max(x))**phantom_param

    elif phantom.lower() == "square":
        if phantom_param is None: phantom_param = 15
        if phantom_param < 3: raise ValueError("The 'square' phantom_param must be larger than or equal to 3")
        x = np.zeros(dim)
        dimh = int(np.round(dim/2))
        w = int(np.round(dim/phantom_param))
        x[(dimh-w):(dimh+w)] = 1
        return x

    elif phantom.lower() == "hat":
        if phantom_param is None: phantom_param = 15
        if phantom_param < 3: raise ValueError("The 'hat' phantom_param must be larger than or equal to 3")
        x = np.zeros(dim)
        dimh = int(np.round(dim/2))
        w = int(np.round(dim/phantom_param))
        x[(dimh-w-1):(dimh)] = np.arange(w+1)/w
        x[(dimh-1):(dimh+w)] = np.flipud(np.arange(w+1))/w
        return x

    elif phantom.lower() == "bumps":
        if phantom_param is not None: warnings.warn("phantom_param is not used for phantom = 'bumps'")
        h = np.pi/dim
        a1 =   1; c1 = 12; t1 =  0.8
        a2 = 0.5; c2 =  5; t2 = -0.5
        grid = np.linspace(0.5,dim-0.5,dim)
        x = a1*np.exp(-c1*(-np.pi/2 + grid*h - t1)**2)+a2*np.exp(-c2*(-np.pi/2 + grid*h - t2)**2)
        return x

    elif phantom.lower() == "derivgauss":
        if phantom_param is None: phantom_param = 5
        x = np.diff(_getExactSolution(dim+1,'gauss',phantom_param))
        return x/np.max(x)
    
    elif phantom.lower() == 'pc':
        if phantom_param is not None: warnings.warn("phantom_param is not used for phantom = 'pc'")
        mesh = np.linspace(0,1,dim)
        x_min, x_max = mesh[0], mesh[-1]
        vals = np.array([0, 2, 3, 2, 0, 1, 0])
        conds = lambda x: [(x_min <= x) & (x < 0.1), (0.1 <= x) & (x < 0.15), (0.15 <= x) & (x < 0.2),  \
                (0.20  <= x) & (x < 0.25), (0.25 <= x) & (x < 0.3), (0.3 <= x) & (x < 0.6), \
                (0.6 <= x) & (x <= x_max)]
        f_signal = lambda x: np.piecewise(x, conds(x), vals)
        x = f_signal(mesh)
        return x

    elif phantom.lower() == 'skyscraper':
        if phantom_param is not None: warnings.warn("phantom_param is not used for phantom = 'skyscraper'")
        x_min = 0 
        x_max = 1
        vals = np.array([0, 1.5, 0, 1.3, 0, 0.75, 0, 0.25, 0, 1, 0])
        conds = lambda x: [(x_min <= x) & (x < 0.10), (0.10 <= x) & (x < 0.15), (0.15 <= x) & (x < 0.20),  \
                        (0.20  <= x) & (x < 0.25), (0.25 <= x) & (x < 0.35), (0.35 <= x) & (x < 0.38),\
                        (0.38  <= x) & (x < 0.45), (0.45 <= x) & (x < 0.55), \
                        (0.55  <= x) & (x < 0.75), (0.75 <= x) & (x < 0.8), (0.8 <= x) & (x <= x_max)]
        f_fun = lambda x: np.piecewise(x, conds(x), vals)
        mesh = np.linspace(x_min, x_max, dim)
        x = f_fun(mesh)
        return x


    else:
        raise NotImplementedError("This phantom is not implemented")


class Poisson1D(BayesianProblem):
    """
    1D Poisson test problem. Discretized 1D Poisson equation (steady-state linear PDE).

    Parameters
    ------------
    dim : int
        | size of the grid for the poisson problem

    endpoint : float
        | Location of end-point of grid.
    
    source : lambda function
        | Function for source term.

    field_type : str or cuqi.geometry.Geometry
        | Field type of domain. The accepted values are:
        | a Geometry object.
        | "KL": a :class:`cuqi.geometry.KLExpansion` geometry object will be created and set as a domain geometry.
        | "KL_Full": a :class:`cuqi.geometry.KLExpansion_Full` geometry object will be created and set as a domain geometry.
        | "Step": a :class:`cuqi.geometry.StepExpansion` geometry object will be created and set as a domain geometry.
        | "CustomKL": a :class:`cuqi.geometry.CustomKL` geometry object will be created and set as a domain geometry.
        | None: a :class:`cuqi.geometry.Continuous1D` geometry object will be created and set as a domain geometry.

    field_params : dict
        | A dictionary of key word arguments that the underlying geometry accepts. (Passed to the underlying geometry when field type is `KL`, `KL_Full`, `CustomKL`, `Step`). For example, for `Step` field type, the dictionary can be `{"n_steps": 3}`.

    map : lambda function
        | Mapping used to modify field.

    imap : lambda function
        | Inverse of KL map.

    SNR : int
        | Signal-to-noise ratio


    observation_grid_map : lambda function
        | Function that takes the grid as input and returns a sub-grid of the nodes where observations are available, e.g. `observation_grid_map = lambda x: x[np.where(x>5.0)]`. 

    Attributes
    ----------
    data : ndarray
        Generated (noisy) data

    model : cuqi.model.PDEModel
        Poisson 1D model

    prior : cuqi.distribution.Distribution
        Distribution of the prior

    likelihood : cuqi.likelihood.Likelihood
        Likelihood function

    exactSolution : ndarray
        Exact solution (ground truth)

    exactData : ndarray
        Noise free data   

    Methods
    ----------
    MAP()
        Compute MAP estimate of posterior.
        NB: Requires prior to be defined.

    sample_posterior(Ns)
        Sample Ns samples of the posterior.
        NB: Requires prior to be defined.

    """
    def __init__(self, dim=128, endpoint=1, source=lambda xs: 10*np.exp( -( (xs - 0.5)**2 ) / 0.02), field_type=None, field_params=None, map=None, imap=None, SNR=200, observation_grid_map=None, exactSolution=None):

        # Prepare PDE form
        N = dim-1   # Number of solution nodes
        dx = endpoint/N   # step size
        grid = np.linspace(dx, endpoint, N, endpoint=False)
        Dx = - np.diag(np.ones(N), 0) + np.diag(np.ones(N-1), 1) #Dx
        vec = np.zeros(N)
        vec[0] = 1
        Dx = np.concatenate([vec.reshape([1, -1]), Dx], axis=0)
        Dx /= dx # FD derivative matrix
        rhs = source(grid)
        
        # Grids for model
        grid_domain = np.linspace(0, endpoint, dim, endpoint=True)
        grid_range  = np.linspace(1./(dim-1), endpoint, dim-1, endpoint=False)

        # PDE form: LHS(x)u=rhs(x)
        grid_obs = grid_range
        if observation_grid_map is not None:
            grid_obs = observation_grid_map(grid_range)
        PDE_form = lambda x: (Dx.T @ np.diag(x) @ Dx, rhs)
        PDE = cuqi.pde.SteadyStateLinearPDE(PDE_form, grid_sol=grid_range,  grid_obs=grid_obs)

        # Set up geometries for model
        if field_params is None:
            field_params = {}
        if isinstance(field_type,Geometry):
            domain_geometry = field_type
        elif field_type=="KL":
            domain_geometry = KLExpansion(grid_domain, **field_params)
        elif field_type=="KL_Full":
            domain_geometry = KLExpansion_Full(grid_domain, **field_params)
        elif field_type=="Step":
            domain_geometry = StepExpansion(grid_domain, **field_params)
        elif field_type=="CustomKL":
            domain_geometry = CustomKL(grid_domain, **field_params)
        else:
            domain_geometry = Continuous1D(grid_domain)

        if map is not None:
            domain_geometry = MappedGeometry(domain_geometry, map, imap)

        range_geometry = Continuous1D(grid_obs)

        # Prepare model
        model = cuqi.model.PDEModel(PDE,range_geometry,domain_geometry)

        # Set up exact solution
        if exactSolution is None:
            x_exact = np.exp( 5*grid_domain*np.exp(-2*grid_domain)*np.sin(endpoint-grid_domain) ) 
        else: 
            x_exact = exactSolution 
        x_exact = CUQIarray(x_exact, is_par=False, geometry=domain_geometry)

        # Generate exact data
        y_exact = model.forward(x_exact,is_par=False)

        # Add noise to data
        sigma = np.linalg.norm(y_exact)/SNR
        sigma2 = sigma*sigma # variance of the observation Gaussian noise
        data = y_exact + np.random.normal(0, sigma, y_exact.shape)

        # Bayesian model
        x = cuqi.distribution.Gaussian(np.zeros(model.domain_dim), 1)
        y = cuqi.distribution.Gaussian(model, sigma2)

        # Initialize Deconvolution as BayesianProblem problem
        super().__init__(y, x, y=data)

        # Store exact values
        self.exactSolution = x_exact
        self.exactData = y_exact

class Heat1D(BayesianProblem):
    """
    1D Heat test problem. Discretized Heat equation (time-dependent linear PDE).

    Parameters
    ------------
    dim : int
        | size of the grid for the heat problem

    endpoint : float
        | Location of end-point of grid.
    
    max_time : float
        | The last time step.
    
    field_type : str or cuqi.geometry.Geometry
        | Field type of domain. The accepted values are:
        | a Geometry object.
        | "KL": a :class:`cuqi.geometry.KLExpansion` geometry object will be created and set as a domain geometry.
        | "KL_Full": a :class:`cuqi.geometry.KLExpansion_Full` geometry object will be created and set as a domain geometry.
        | "Step": a :class:`cuqi.geometry.StepExpansion` geometry object will be created and set as a domain geometry.
        | "CustomKL": a :class:`cuqi.geometry.CustomKL` geometry object will be created and set as a domain geometry.
        | None: a :class:`cuqi.geometry.Continuous1D` geometry object will be created and set as a domain geometry.

    field_params : dict
        | A dictionary of key word arguments that the underlying geometry accepts. (Passed to the underlying geometry when field type is `KL`, `KL_Full`, `CustomKL`, `Step`). For example, for `Step` field type, the dictionary can be `{"n_steps": 3}`.

    map : lambda function
        | If given, an underlying `MappedGeometry` geometry object is created which applies the mapping on the field, e.g. for log parameterization: `map = lambda x:np.exp(x)`.

    imap : lambda function
        | The inverse of the provided map. 

    SNR : int
        | Signal-to-noise ratio

    exactSolution : ndarray or CUQIarray
        | The exact solution of the problem which is the heat model initial condition in this test problem. If provided as None, an exact solution is generated, if provided as ndarray, it is assumed to be function values and it is converted to a CUQIarray.


    observation_grid_map : lambda function
       | Function that takes the grid as input and returns a sub-grid of the nodes where observations are available, e.g. `observation_grid_map = lambda x: x[np.where(x>5.0)]`. 
 
    Attributes
    ----------
    data : ndarray
        Generated (noisy) data

    model : cuqi.model.PDEModel
        Heat 1D model

    prior : cuqi.distribution.Distribution
        Distribution of the prior

    likelihood : cuqi.likelihood.Likelihood
        Likelihood function

    exactSolution : ndarray
        Exact solution (ground truth)

    exactData : ndarray
        Noise free data   

    Methods
    ----------
    MAP()
        Compute MAP estimate of posterior.
        NB: Requires prior to be defined.

    sample_posterior(Ns)
        Sample Ns samples of the posterior.
        NB: Requires prior to be defined.

    """
    def __init__(self, dim=128, endpoint=1, max_time=0.2, field_type=None, field_params=None,map=None, imap=None, SNR=200, exactSolution=None, observation_grid_map=None):
        

        # Prepare PDE form
        N = dim   # Number of solution nodes
        dx = endpoint/(N+1)   # space step size
        cfl = 5/11 # the cfl condition to have a stable solution
        dt_approx = cfl*dx**2 # defining approximate time step size
        max_iter = int(max_time/dt_approx) # number of time steps
        Dxx = (np.diag( -2*np.ones(N) ) + np.diag(np.ones(N-1),-1) + np.diag(np.ones(N-1),1))/dx**2 # FD diffusion operator
        
        # Grids for model
        grid_domain = np.linspace(dx, endpoint, N, endpoint=False)
        grid_range = np.linspace(dx, endpoint, N, endpoint=False) 
        time_steps = np.linspace(0,max_time,max_iter+1,endpoint=True)

        # PDE form (diff_op, IC, time_steps)
        grid_obs = grid_range
        if observation_grid_map is not None:
            grid_obs = observation_grid_map(grid_obs)

        def PDE_form(IC, t): return (Dxx, np.zeros(N), IC)
        PDE = cuqi.pde.TimeDependentLinearPDE(
            PDE_form, time_steps, grid_sol=grid_domain, grid_obs=grid_obs)

        # Set up geometries for model
        if field_params is None:
            field_params = {}
        if isinstance(field_type,Geometry):
            domain_geometry = field_type
        elif field_type=="KL":
            domain_geometry = KLExpansion(grid_domain, **field_params)
        elif field_type=="KL_Full":
            domain_geometry = KLExpansion_Full(grid_domain, **field_params)
        elif field_type=="Step":
            domain_geometry = StepExpansion(grid_domain, **field_params)
        elif field_type=="CustomKL":
            domain_geometry = CustomKL(grid_domain, **field_params)
        else:
            domain_geometry = Continuous1D(grid_domain)
        if map is not None:
            domain_geometry = MappedGeometry(domain_geometry,map,imap)
        range_geometry = Continuous1D(grid_obs)

        # Prepare model
        model = cuqi.model.PDEModel(PDE,range_geometry,domain_geometry)
        if exactSolution is not None:
            x_exact = CUQIarray(exactSolution, is_par = False, geometry=domain_geometry)
        # Set up exact solution
        
        else:
            if field_type=="Step":
                n_steps = domain_geometry.n_steps
                x_exact = CUQIarray(domain_geometry.par2fun(np.array(range(n_steps))), is_par=False, geometry=domain_geometry)
            else:
                grid_domain = model.domain_geometry.grid
                x_exact = grid_domain*np.exp(-2*grid_domain)*np.sin(endpoint-grid_domain)
                x_exact = CUQIarray(x_exact, is_par=False, geometry=domain_geometry)
        #x_exact = 100*grid_domain*np.exp(-5*grid_domain)*np.sin(endpoint-grid_domain)
        # Generate exact data
        y_exact = model.forward(x_exact, is_par=False)
        # Add noise to data
        sigma = np.linalg.norm(y_exact)/SNR
        sigma2 = sigma*sigma # variance of the observation Gaussian noise
        data = y_exact + np.random.normal(0, sigma, y_exact.shape)

        # Bayesian model
        x = cuqi.distribution.Gaussian(np.zeros(model.domain_dim), 1)
        y = cuqi.distribution.Gaussian(model(x), sigma2)
        
        # Initialize Deconvolution as BayesianProblem problem
        super().__init__(y, x, y=data)

        # Store exact values
        self.exactSolution = x_exact
        self.exactData = y_exact
        self.infoString = f"Noise type: Additive i.i.d. noise with mean zero and signal to noise ratio: {SNR}"


class Abel1D(BayesianProblem):
    """
    1D Abel test problem. 1D model of rotationally symmetric computed tomography.

    Parameters
    ------------
    dim : int
        size of the grid for the problem

    endpoint : float
        Location of end-point of grid.
    
    field_type : str or cuqi.geometry.Geometry
        Field type of domain.

    KL_map : lambda function
        Mapping used to modify field.

    KL_imap : lambda function
        Inverse of KL map.

    SNR : int
        Signal-to-noise ratio

    Attributes
    ----------
    data : ndarray
        Generated (noisy) data

    model : cuqi.model.LinearModel
        Abel 1D model

    prior : cuqi.distribution.Distribution
        Distribution of the prior

    likelihood : cuqi.likelihood.Likelihood
        Likelihood function

    exactSolution : ndarray
        Exact solution (ground truth)

    exactData : ndarray
        Noise free data   

    Methods
    ----------
    MAP()
        Compute MAP estimate of posterior.
        NB: Requires prior to be defined.

    sample_posterior(Ns)
        Sample Ns samples of the posterior.
        NB: Requires prior to be defined.

    """
    def __init__(self, dim=128, endpoint=1, field_type=None, field_params=None, KL_map=None, KL_imap=None, SNR=100):
        N = dim # number of quadrature points
        h = endpoint/N # quadrature weight

        tvec = np.linspace(h/2, endpoint-h/2, N).reshape(1, -1) 
        svec = tvec.reshape(-1, 1) + h/2
        tmat = np.tile( tvec, [N, 1] )
        smat = np.tile( svec, [1, N] )
        
        idx = np.where(tmat<smat) # only applying the quadrature on 0<x<1
        A = np.zeros([N,N]) # Abel integral operator
        A[idx[0], idx[1]] = h/np.sqrt( np.abs( smat[idx[0], idx[1]] - tmat[idx[0], idx[1]] ) )

        # discretization
        grid = np.linspace(0, endpoint, N)

        # Geometry
        if field_params is None:
            field_params = {}
        if isinstance(field_type,Geometry):
            domain_geometry = field_type
        elif field_type=="KL":
            domain_geometry = KLExpansion(grid,**field_params)
        elif field_type=="Step":
            domain_geometry = StepExpansion(grid,**field_params)
        elif field_type=="CustomKL":
            domain_geometry = CustomKL(grid,**field_params)
        else:
            domain_geometry = Continuous1D(grid)

        if KL_map is not None:
            domain_geometry = MappedGeometry(domain_geometry,KL_map,KL_imap)

        range_geometry = Continuous1D(grid)

        # Set up model
        model = LinearModel(A,range_geometry=range_geometry, domain_geometry=domain_geometry)
    
        # Set up exact solution
        x_exact = np.sin(tvec*np.pi)*np.exp(-2*tvec)
        x_exact.shape = (dim,)
        x_exact = CUQIarray(x_exact, is_par=False, geometry=domain_geometry)

        # Generate exact data
        y_exact = model.forward(x_exact,is_par=False)

        # Add noise to data
        sigma = np.linalg.norm(y_exact)/SNR
        sigma2 = sigma*sigma # variance of the observation Gaussian noise
        data = y_exact + np.random.normal(0, sigma, y_exact.shape )

        # Bayesian model
        x = cuqi.distribution.Gaussian(np.zeros(model.domain_dim), 1)
        y = cuqi.distribution.Gaussian(model(x), sigma2)
        
        # Initialize Deconvolution as BayesianProblem problem
        super().__init__(y, x, y=data)

        # Store exact values
        self.exactSolution = x_exact
        self.exactData = y_exact


class _Deconv_1D(BayesianProblem):
    """
    1D Deconvolution test problem. Discreate linear problem from blurring kernel.

    Parameters
    ------------
    dim : int
        size of the grid for the problem

    endpoint : float
        Location of end-point of grid.
    
    field_type : str or cuqi.geometry.Geometry
        Field type of domain.

    KL_map : lambda function
        Mapping used to modify field.

    KL_imap : lambda function
        Inverse of KL map.

    SNR : int
        Signal-to-noise ratio

    Attributes
    ----------
    data : ndarray
        Generated (noisy) data

    model : cuqi.model.LinearModel
        Deconvolution 1D model

    prior : cuqi.distribution.Distribution
        Distribution of the prior

    likelihood : cuqi.likelihood.Likelihood
        Likelihood function

    exactSolution : ndarray
        Exact solution (ground truth)

    exactData : ndarray
        Noise free data   

    Methods
    ----------
    MAP()
        Compute MAP estimate of posterior.
        NB: Requires prior to be defined.

    sample_posterior(Ns)
        Sample Ns samples of the posterior.
        NB: Requires prior to be defined.

    """
    def __init__(self, dim=128, endpoint=1, kernel=None, blur_size=48, field_type=None, field_params=None, KL_map=None, KL_imap=None, SNR=100):
        
        warnings.warn("DEPRECATED: Use Deconvolution1D instead.")
        
        N = dim # number of quadrature points
        h = endpoint/N # quadrature weight
        grid = np.linspace(0, endpoint, N)

        if kernel is None:
            kernel = lambda x, y, blur_size_var: blur_size_var / 2*np.exp(-blur_size*abs((x-y)))   # blurring kernel
        
        # convolution matrix
        T1, T2 = np.meshgrid(grid, grid)
        A = h*kernel(T1, T2, blur_size)
        maxval = A.max()
        A[A < 5e-3*maxval] = 0
        A = csc_matrix(A)   # make A sparse
        
        # discretization
        if isinstance(field_type,Geometry):
            domain_geometry = field_type
        elif field_type=="KL":
            domain_geometry = KLExpansion(grid,field_params)
        elif field_type=="Step":
            domain_geometry = StepExpansion(grid)
        elif field_type=="CustomKL":
            domain_geometry = CustomKL(grid,field_params)
        else:
            domain_geometry = Continuous1D(grid)

        if KL_map is not None:
            domain_geometry = MappedGeometry(domain_geometry,KL_map,KL_imap)

        range_geometry = Continuous1D(grid)

        # Set up model
        model = LinearModel(A,range_geometry=range_geometry, domain_geometry=domain_geometry)
    
        # Prior
        prior = cuqi.distribution.Gaussian(np.zeros(model.domain_dim), 1, geometry=model.domain_geometry, name="x")

        # Set up exact solution
        x_exact = prior.sample()

        # Generate exact data
        y_exact = model.forward(x_exact)

        # Add noise to data
        sigma = np.linalg.norm(y_exact)/SNR
        sigma2 = sigma*sigma # variance of the observation Gaussian noise
        data = y_exact + np.random.normal(0, sigma, y_exact.shape)

        likelihood = cuqi.distribution.Gaussian(model(prior), sigma2, name="y").to_likelihood(data)
        
        # Initialize Deconvolution as BayesianProblem problem
        super().__init__(likelihood, prior)

        # Store exact values
        self.exactSolution = x_exact
        self.exactData = y_exact


#=============================================================================
class Deconvolution2D(BayesianProblem):
    """
    
    Create a 2D Deconvolution test problem defined by the inverse problem

    .. math::

        \mathbf{b} = \mathbf{A}\mathbf{x},

    where :math:`\mathbf{b}` is a (noisy) blurred image,
    :math:`\mathbf{x}` is a sharp image and
    :math:`\mathbf{A}` is a convolution operator.

    The convolution operator is defined by specifying a point spread function and
    boundary conditions and is computed (matrix-free) via scipy.signal.fftconvolve.
    The inputs are padded to fit the boundary conditions.
    https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.fftconvolve.html.

    Parameters
    ----------
    dim : int
        | size of the (dim,dim) deconvolution problem

    PSF : string or ndarray
        | Determines type of the underlying point spread function (PSF).
        | 'Gauss' - a Gaussian blur
        | 'Moffat' - a Moffat blur blur
        | 'Defocus' - an out-of-focus blur
        | ndarray - Custom user-specified PSF

    PSF_param : scalar
        | A parameter that determines the shape of the PSF;
        | the larger the parameter, the larger blur on the image.
        | Ignored if PSF is given as ndarray

    PSF_size : int
        | Defines the size of the PSF. The size becomes (PSF_size, PSF_size).
        | Ignored if PSF is given as ndarray

    BC : string
        | Boundary conditions of convolution
        | 'zero' - Zero boundary
        | 'periodic' - Periodic boundary
        | 'Neumann' - Neumann (reflective) boundary
        | 'Mirror' - Mirrored boundary
        | 'Nearest' - Replicates last element of boundary
        
    phantom : string or ndarray
        | The phantom (sharp image) that is convolved.
        | If ndarray it should be a square 2D array representing an image.
        | The image will automatically be resized to fit the problem size.
        | If string it should be any 2D phantom defined in cuqi.data.
        | The string is lowercased and any hyphens are replaced 
        | with underscores to match a method name in cuqi.data.
        | Examples:
        | 'astronaut' - a photo of an astronaut.
        | 'camera' - a photo of a man with a camera.
        | 'cat' - a photo of a cat.
        | 'cookie' - cartoon cookie.
        | 'satellite' - a photo of a satellite.

    noise_type : string
        | The type of noise
        | "Gaussian" - Gaussian white noise¨
        | "scaledGaussian" - Gaussian noise where standard deviation is scaled by by data.

    noise_std : scalar
        | Standard deviation of the noise.

    prior : Distribution
        | Option to set the prior distribution.
        | If set all components of the posterior
        | are defined and sampling can be achieved.

    Attributes
    ----------
    data : CUQIarray
        Generated (noisy) data

    model : Model
        Deconvolution forward model

    prior : Distribution
        Distribution of the prior

    likelihood : Likelihood
        The likelihood function
        (automatically computed from data distribution)

    exactSolution : CUQIarray
        Exact solution (ground truth)

    exactData : CUQIarray
        Noise free data.

    Methods
    -------
    MAP()
        Compute MAP estimate of posterior.
        NB: Requires prior to be defined

    sample_posterior(Ns)
        Sample Ns samples of the posterior.
        NB: Requires prior to be defined
    """
    def __init__(self,
        dim=128,
        PSF="gauss",
        PSF_param=2.56,
        PSF_size=21,
        BC="periodic",
        phantom="satellite",
        noise_type="gaussian",
        noise_std=0.0036,
        prior = None):
        
        # setting up the geometry
        domain_geometry = Image2D((dim, dim))
        range_geometry = Image2D((dim, dim))

        # set boundary conditions
        if (BC.lower() == "neumann"):
            BC = "symmetric"
        elif (BC.lower() == "zero"):
            BC = "constant" # cval = 0
        elif (BC.lower() == "nearest"):
            BC = "edge" # the input is extended by replicating the last pixel
        elif (BC.lower() == "mirror"):
            BC = "reflect" # reflecting about the center of the last pixel
        elif (BC.lower() == "periodic"):
            BC = "wrap"
        else:
            raise TypeError("Unknown BC type")

        # Set up PSF.
        if isinstance(PSF, np.ndarray):
            P = PSF
        elif isinstance(PSF, str):
            if PSF.lower() == "gauss":
                P, _ = _GaussPSF(np.array([PSF_size, PSF_size]), PSF_param)
            elif PSF.lower() == "moffat":
                P, _ = _MoffatPSF(np.array([PSF_size, PSF_size]), PSF_param, 1)
            elif PSF.lower() == "defocus":
                P, _ = _DefocusPSF(np.array([PSF_size, PSF_size]), PSF_param)
        else:
            raise TypeError(f"Unknown PSF: {PSF}.")

        # build forward model
        model = cuqi.model.LinearModel(lambda x: _proj_forward_2D(x, P, BC), 
                                       lambda x: _proj_backward_2D(x, P, BC), 
                                       range_geometry, 
                                       domain_geometry)

        # User provided phantom as ndarray
        if isinstance(phantom, np.ndarray):
            if phantom.ndim > 2:
                raise ValueError("Input phantom image must be an image (matrix) or vector")
            if phantom.ndim == 1:
                N = int(round(np.sqrt(len(phantom))))
                phantom = phantom.reshape(N,N)
            phantom = cuqi.data.imresize(phantom, dim) # Resize phantom (if wrong size)
            x_exact2D = phantom
        # If phantom is string its a specific case
        elif isinstance(phantom, str):
            # lowercase and replace hyphens with underscores to match library method names
            phantom = phantom.lower().replace("-", "_") 
            if hasattr(cuqi.data, phantom):
                x_exact2D = getattr(cuqi.data, phantom)(size=dim)
            else:
                raise ValueError(f"Phantom {phantom} not found in cuqi.data phantom library.")
        else:
            raise TypeError("Unknown phantom type. Must be ndarray or string.")

        x_exact = x_exact2D.flatten()
        x_exact = CUQIarray(x_exact, is_par=True, geometry=domain_geometry)

        # Generate exact data (blurred)
        y_exact = model@x_exact

        # Create prior
        if prior is None:
            prior = cuqi.distribution.Gaussian(np.zeros(model.domain_dim), 1, geometry=domain_geometry, name="x")

        # Data distribution
        if noise_type.lower() == "gaussian":
            data_dist = cuqi.distribution.Gaussian(model(prior), noise_std**2, geometry=range_geometry, name="y")
        elif noise_type.lower() == "scaledgaussian":
            data_dist = cuqi.distribution.Gaussian(model(prior), (y_exact*noise_std)**2, geometry=range_geometry, name="y")
        else:
            raise NotImplementedError("This noise type is not implemented")
        
        # Generate noisy data
        data = data_dist(x_exact).sample()

        # Likelihood
        likelihood = data_dist.to_likelihood(data)
        
        # Initialize Deconvolution as BayesianProblem problem
        super().__init__(likelihood, prior)

        self.exactSolution = x_exact
        self.exactData = y_exact
        self.infoString = "Noise type: Additive {} with std: {}".format(noise_type.capitalize(),noise_std)
        self.Miscellaneous = {"PSF" : P, "BC": BC}

#=========================================================================
def _proj_forward_2D(X, P, BC):
    PSF_size = max(P.shape)
    X_padded = np.pad(X, PSF_size//2, mode=BC)
    Ax = fftconvolve(X_padded, P, mode='valid')
    if not PSF_size & 0x1: # If PSF_size is even
        Ax = Ax[1:, 1:] # Remove first row and column to fit convolve math
    return Ax

#=========================================================================
def _proj_backward_2D(B, P, BC):
    P = np.flipud(np.fliplr(P)) # Flip PSF
    return _proj_forward_2D(B, P, BC)

# ===================================================================
# Array with PSF for Gaussian blur (astronomic turbulence)
# ===================================================================
def _GaussPSF(dim, s):
    if hasattr(dim, "__len__"):
        m, n = dim[0], dim[1]
    else:
        m, n = dim, dim
    s1, s2 = s, s
    
    # Set up grid points to evaluate the Gaussian function
    x = np.arange(-np.fix(n/2), np.ceil(n/2))
    y = np.arange(-np.fix(m/2), np.ceil(m/2))
    X, Y = np.meshgrid(x, y)

    # Compute the Gaussian, and normalize the PSF.
    PSF = np.exp( -0.5* ((X**2)/(s1**2) + (Y**2)/(s2**2)) )
    PSF /= PSF.sum()

    # find the center
    mm, nn = np.where(PSF == PSF.max())
    center = np.array([mm[0], nn[0]])

    return PSF, center.astype(int)
# ===================================================================
# Array with PSF for Moffat blur (astronomical telescope)
# ===================================================================
def _MoffatPSF(dim, s, beta):
    if hasattr(dim, "__len__"):
        m, n = dim[0], dim[1]
    else:
        m, n = dim, dim
    s1, s2 = s, s
    
    # Set up grid points to evaluate the Gaussian function
    x = np.arange(-np.fix(n/2), np.ceil(n/2))
    y = np.arange(-np.fix(m/2), np.ceil(m/2))
    X, Y = np.meshgrid(x, y)

    # Compute the Gaussian, and normalize the PSF.
    PSF = ( 1 + (X**2)/(s1**2) + (Y**2)/(s2**2) )**(-beta)
    PSF = PSF / PSF.sum()

    # find the center
    mm, nn = np.where(PSF == PSF.max())
    center = np.array([mm[0], nn[0]])

    return PSF, center.astype(int)
# ===================================================================
# Array with PSF for out-of-focus blur
# ===================================================================
def _DefocusPSF(dim, R):
    if hasattr(dim, "__len__"):
        m, n = dim[0], dim[1]
    else:
        m, n = dim, dim
    
    center = np.fix((np.array([m, n]))/2)
    if (R == 0):    
        # the PSF is a delta function and so the blurring matrix is I
        PSF = np.zeros((m, n))
        PSF[center[0], center[1]] = 1
    else:
        PSF = np.ones((m, n)) / (np.pi * R**2)
        k = np.arange(1, max(m, n)+1)
        aa, bb = (k-center[0])**2, (k-center[1])**2
        A, B = np.meshgrid(aa, aa), np.meshgrid(bb, bb)
        idx = np.array(((A[0].T + B[0]) > (R**2)))
        PSF[idx] = 0
    PSF = PSF / PSF.sum()

    return PSF, center.astype(int)


#=============================================================================
class WangCubic(BayesianProblem):
    """ Two parameters and one observation cubic test problem.
    
    Parameters
    ------------
    noise_std : scalar
        Standard deviation of the noise

    prior : cuqi.distribution.Distribution
        Distribution of the prior
    
    data : scalar
        Observed data
    
    Notes
    -----
    Based on Section 3.3.2 in Wang (2015):
    Z. Wang, "An Optimization Based Algorithm for Bayesian Inference". Master thesis. MIT. 2015  
    https://dspace.mit.edu/bitstream/handle/1721.1/98815/921147308-MIT.pdf?sequence=1&isAllowed=y

    """
    def __init__(self, noise_std=1, prior=None, data=None):
        # forward model and gradient
        def forward(x):
            return 10*x[1] - 10*x[0]**3 + 5*x[0]**2 + 6*x[0]
        def jacobian(x):
            return np.array([[-30*x[0]**2 + 10*x[0] + 6, 10]])
        model = cuqi.model.Model(forward, range_geometry=1, domain_geometry=2, jacobian=jacobian)

        # define prior
        if prior is None:
            prior = cuqi.distribution.Gaussian(np.array([1, 0]), 1, name="x")

        # data
        if data is None:
            data = 1

        # data distribution is Gaussian
        data_dist = cuqi.distribution.Gaussian(model(prior), noise_std**2, name="y")

        # Define Gaussian likelihood
        likelihood = data_dist.to_likelihood(data)

        # Set up Bayesian Problem
        super().__init__(likelihood, prior)

        # Store exact values
        self.exactSolution = None
        self.exactData = None
        self.infoString = "Noise type: Additive {} with std: {}".format('Gaussian', noise_std)



import pytest
from cuqi.distribution import LMRF, CMRF, GMRF
from cuqi.geometry import Image2D, Geometry, _DefaultGeometry2D
import numpy as np

class Image3D(Geometry):
    """ Test class for 3 dimensional geometry """
    def __init__(self):
        pass
    @property
    def par_shape(self):
        return (8**3,)
    @property
    def fun_shape(self):
        return (8, 8, 8)
    def _plot(self):
        pass
    

@pytest.mark.parametrize("scale_value", [0.1, lambda s: s])
def test_LMRF_should_not_allow_None_or_scalar_geometry(scale_value):
    with pytest.raises(ValueError, match="supported geometry"):
        LMRF(0, scale=scale_value)

@pytest.mark.parametrize("physical_dim", [-1, 0, 3])
def test_LMRF_invalid_physical_dim_raises_error(physical_dim):
    with pytest.raises(ValueError, match="Only physical dimension 1 or 2 supported."):
        LMRF(0, scale=0.2, geometry=Image3D())

def test_LMRF_mismatched_location_and_geometry_raises_error():
    image_2d_geometry = Image2D((8, 8))
    with pytest.raises(TypeError):
        LMRF(np.zeros(3), scale=0.2, geometry=image_2d_geometry)

def test_LMRF_default_geometry_is_replaced_when_physical_dim_is_2():
    lmrf = LMRF(0, scale=0.2, geometry=(4,4))
    assert isinstance(lmrf.geometry, _DefaultGeometry2D)

@pytest.mark.parametrize("scale_value", [0.1, lambda s: s])
def test_CMRF_should_not_allow_None_or_scalar_geometry(scale_value):
    with pytest.raises(ValueError, match="supported geometry"):
        CMRF(0, scale=scale_value)

@pytest.mark.parametrize("physical_dim", [-1, 0, 3])
def test_CMRF_invalid_physical_dim_raises_error(physical_dim):
    with pytest.raises(ValueError, match="Only physical dimension 1 or 2 supported."):
        CMRF(0, scale=0.2, geometry=Image3D())

def test_CMRF_mismatched_location_and_geometry_raises_error():
    image_2d_geometry = Image2D((8, 8))
    with pytest.raises(TypeError):
        CMRF(np.zeros(3), scale=0.2, geometry=image_2d_geometry)

def test_CMRF_default_geometry_is_replaced_when_physical_dim_is_2():
    cmrf = CMRF(0, scale=0.2, geometry=(4,4))
    assert isinstance(cmrf.geometry, _DefaultGeometry2D)

def test_LMRF_dim_vs_ndarray_location():
    lmrf1 = LMRF(0, scale=0.2, geometry=(4,4))
    lmrf2 = LMRF(np.zeros(16), scale=0.2, geometry=Image2D((4,4)))
    assert lmrf1.dim == lmrf2.dim
    assert lmrf1.logd(np.zeros(16)) == lmrf2.logd(np.zeros(16))

def test_CMRF_dim_vs_ndarray_location():
    cmrf1 = CMRF(0, scale=0.2, geometry=(4,4))
    cmrf2 = CMRF(np.zeros(16), scale=0.2, geometry=Image2D((4,4)))
    assert cmrf1.dim == cmrf2.dim
    assert cmrf1.logd(np.zeros(16)) == cmrf2.logd(np.zeros(16))
    assert np.allclose(cmrf1.gradient(np.zeros(16)),cmrf2.gradient(np.zeros(16)))

# GMRF tests
def test_GMRF_should_not_allow_None_or_scalar_geometry():
    with pytest.raises(ValueError, match="supported geometry"):
        GMRF(0, 0.2)

@pytest.mark.parametrize("physical_dim", [-1, 0, 3])
def test_GMRF_invalid_physical_dim_raises_error(physical_dim):
    with pytest.raises(ValueError, match="Only physical dimension 1 or 2 supported."):
        GMRF(0, 0.2, geometry=Image3D())

def test_GMRF_mismatched_location_and_geometry_raises_error():
    image_2d_geometry = Image2D((8, 8))
    with pytest.raises(TypeError):
        GMRF(np.zeros(3), 0.2, geometry=image_2d_geometry)

def test_GMRF_default_geometry_is_replaced_when_physical_dim_is_2():
    gmrf = GMRF(0, 0.2, geometry=(4,4))
    assert isinstance(gmrf.geometry, _DefaultGeometry2D)

def test_GMRF_dim_vs_ndarray_location():
    gmrf1 = GMRF(0, 0.2, geometry=(4,4))
    gmrf2 = GMRF(np.zeros(16), 0.2, geometry=Image2D((4,4)))
    assert gmrf1.dim == gmrf2.dim
    assert gmrf1.logd(np.zeros(16)) == gmrf2.logd(np.zeros(16))
    assert np.allclose(gmrf1.gradient(np.zeros(16)),gmrf2.gradient(np.zeros(16)))





import cuqi
import pytest
import numpy as np
import scipy.sparse as sps
from pytest import approx

def test_likelihood_log_and_grad():
    #Create likelihood
    model, data, probInfo = cuqi.testproblem.Deconvolution1D().get_components()
    likelihood = cuqi.distribution.Gaussian(model,1).to_likelihood(data)

    # Tests log and gradient calls do not cause errors
    likelihood.logd(probInfo.exactSolution)
    likelihood.gradient(probInfo.exactSolution)

def test_likelihood_attributes():
    model, data, _ = cuqi.testproblem.Poisson1D(dim=128, field_type="Step").get_components()
    likelihood = cuqi.distribution.Gaussian(model,1).to_likelihood(data)

    # dim of domain
    assert likelihood.dim == model.domain_dim

    # Geometry of domain
    assert likelihood.geometry == model.domain_geometry

    # Shape of domain
    assert likelihood.par_shape == model.domain_geometry.par_shape
    assert likelihood.fun_shape == model.domain_geometry.fun_shape

    # Model extraction
    assert likelihood.model == model

@pytest.mark.parametrize("dist",[
                        # ------------ Scalar --------------
                        cuqi.distribution.Gaussian(np.zeros(128), sqrtcov=np.pi),
                        cuqi.distribution.Gaussian(np.zeros(128), cov=np.pi**2),
                        cuqi.distribution.Gaussian(np.zeros(128), prec=1/np.pi**2),
                        cuqi.distribution.Gaussian(np.zeros(128), sqrtprec=1/np.pi),
                        # ------------ VECTOR --------------
                        cuqi.distribution.Gaussian(np.zeros(128), sqrtcov=np.pi*np.ones(128)),
                        cuqi.distribution.Gaussian(np.zeros(128), cov=(np.pi**2)*np.ones(128)),
                        cuqi.distribution.Gaussian(np.zeros(128), prec=1/(np.pi**2)*np.ones(128)),
                        cuqi.distribution.Gaussian(np.zeros(128), sqrtprec=1/np.pi*np.ones(128)),
                        # ------------ Diagonal matrix --------------
                        cuqi.distribution.Gaussian(np.zeros(128), sqrtcov=np.pi*np.eye(128)),
                        cuqi.distribution.Gaussian(np.zeros(128), cov=(np.pi**2)*np.eye(128)),
                        cuqi.distribution.Gaussian(np.zeros(128), prec=1/(np.pi**2)*np.eye(128)),
                        cuqi.distribution.Gaussian(np.zeros(128), sqrtprec=1/np.pi*np.eye(128)),
                        # ------------ Sparse diagonal matrix --------------
                        cuqi.distribution.Gaussian(np.zeros(128), sqrtcov=np.pi*sps.eye(128)),
                        cuqi.distribution.Gaussian(np.zeros(128), cov=(np.pi**2)*sps.eye(128)),
                        cuqi.distribution.Gaussian(np.zeros(128), prec=1/(np.pi**2)*sps.eye(128)),
                        cuqi.distribution.Gaussian(np.zeros(128), sqrtprec=1/np.pi*sps.eye(128)),
                        ])
def test_likelihood_Gaussian_log_IID(dist):
    model, data, probInfo = cuqi.testproblem.Deconvolution1D(dim=128).get_components()
    dist.mean = model
    likelihood = dist.to_likelihood(probInfo.exactData) #We use exact data to get same logpdf every time
    assert likelihood.logd(probInfo.exactSolution) == approx(-264.14955763892135)

def test_likelihood_UserDefined():
    # CUQI likelihood
    model, data, probInfo = cuqi.testproblem.Deconvolution1D().get_components()
    L = cuqi.distribution.Gaussian(model, 1).to_likelihood(data)

    # Create user defined likelihood
    likelihood = cuqi.likelihood.UserDefinedLikelihood(dim=L.dim, logpdf_func=L.logd, gradient_func=L.gradient, geometry=L.geometry)
    
    # log
    assert likelihood.logd(probInfo.exactSolution) == L.logd(probInfo.exactSolution)

    # gradient
    assert np.allclose(likelihood.gradient(probInfo.exactSolution), L.gradient(probInfo.exactSolution))

    # dim
    assert likelihood.dim == model.domain_dim

    # geometry
    assert likelihood.geometry == model.domain_geometry

@pytest.mark.parametrize("dist",[
    cuqi.distribution.Gaussian(geometry=2),
    cuqi.distribution.Gaussian(lambda x: x, lambda s: s, geometry=2),
    cuqi.distribution.Gaussian(cuqi.model.Model(lambda x: x, 2, 2), lambda s:s, geometry=2)
])
@pytest.mark.parametrize("mean, cov, data",[
    (np.zeros(2), np.eye(2), np.ones(2)),
    (np.ones(2), 0.1*np.eye(2), np.random.rand(2)),
])
def test_likelihood_conditioning(dist, mean, cov, data):
    """ Test conditioning on parameters of likelihood for Gaussian """

    # Create likelihood
    likelihood = dist.to_likelihood(data)

    # Get parameter names
    param_names = likelihood.get_parameter_names()

    # Create dict for mean and cov (assumes mean is first parameter, which is not always true)
    mean_dict = {param_names[0]: mean}
    cov_dict = {param_names[1]: cov}

    # Full param dict
    param_dict = {**mean_dict, **cov_dict}

    # Evaluate logp from full params
    log_val = likelihood.logd(**param_dict)

    # Now compare log when conditioning on 3 cases: mean, cov, both
    assert likelihood(**mean_dict).logd(**cov_dict) == log_val
    assert likelihood(**cov_dict).logd(**mean_dict) == log_val
    assert likelihood(**param_dict).logd() == log_val # This becomes EvaluatedDensity. Throws warning on the stack size (since we pass name to EvaluatedDensity)


@pytest.mark.parametrize("y",
                         [cuqi.distribution.Gaussian(np.zeros(6), np.eye(6)),
                          cuqi.distribution.Lognormal(np.zeros(6), 4)])
@pytest.mark.parametrize("x_i", [np.array([0.1, 0.3, 6, 12, 1, 2]),
                                 np.array([0.1, 0.3, 0.5, 6, 3, 1])])
def test_enable_FD_gradient(y, x_i):
    """ Test that the likelihood exact gradient and FD gradient are close."""

    # Create a model
    model = cuqi.testproblem.Deconvolution1D(dim=6).model

    # Create likelihood
    y.mean = model
    data = y(x_i).sample()
    likelihood = y(y=data)

    # Compute exact gradient
    g_exact = likelihood.gradient(x_i)

    # Compute FD gradient
    likelihood.enable_FD(1e-7)
    g_FD = likelihood.gradient(x_i)

    # Assert that the exact and FD gradient are close,
    # but not exactly equal (since we use a different method)
    assert np.allclose(g_exact, g_FD) and np.all(g_exact != g_FD)



import numpy as np
from cuqi.distribution import Distribution

class Laplace(Distribution):
    """ Laplace distribution. 

    Defines a Laplace distribution given a location and a scale. The Laplace distribution is defined as

    .. math::

        p(x) = \\frac{1}{2b} \exp\left(-\\frac{|x-\mu|}{b}\\right),

    where :math:`\mu` is the location (mean) and :math:`b` is the scale (decay) parameter.

    The rate parameter is defined as :math:`\lambda = \\frac{1}{b}`.

    The variables of this Laplace distribution are independent identically distributed (i.i.d.).

    Parameters
    ----------
    location : scalar or ndarray
        The location parameter of the distribution.

    scale : scalar
        The scale parameter of the distribution.

    """

    def __init__(self, location, scale, **kwargs):

        # Init from abstract distribution class
        super().__init__(**kwargs)

        self.location = location
        self.scale = scale
  
    def logpdf(self, x):
        if isinstance(x, (float,int)):
            x = np.array([x])
        return self.dim*(np.log(0.5/self.scale)) - np.linalg.norm(x-self.location,1)/self.scale

    def _sample(self,N=1,rng=None):
        if rng is not None:
            s =  rng.laplace(self.location, self.scale, (N,self.dim)).T
        else:
            s = np.random.laplace(self.location, self.scale, (N,self.dim)).T
        return s



import cuqi
import pytest
import numpy as np

def test_density_variable_name_detection():
    """Test that the density variable name is detected correctly at different levels of the python stack. """

    # Test that the density variable name is detected correctly at current level.
    x = cuqi.distribution.Gaussian(geometry=1)
    assert x.name == 'x'

    # Test that variable name is detected correctly 1 level deep.
    def inner_name():
        y = cuqi.distribution.Gaussian(geometry=1)
        assert y.name == 'y'
    inner_name()

    # Test variable name is detected correctly at n levels deep.
    class recursive_name:
        def __init__(self, max_recursion=10):
            self.max_recursion = max_recursion
        def __call__(self, current_recursion=0):
            if current_recursion == self.max_recursion:
                z = cuqi.distribution.Gaussian(geometry=1)
                assert z.name == 'z'
            else:
                self(current_recursion + 1)
    recursive_name()()

def test_variable_name_accross_frames():
    """ Test variable name can be inferred across multiple stack frames. """

    h = cuqi.distribution.Gaussian(geometry=1) # Name should be 'h'

    def recursive_return_dist(dist, recursions):
        if recursions == 0:
            assert dist.name == 'h' # h was defined many frames above, and name should be inferred correctly.
        else:
            recursive_return_dist(dist, recursions - 1)
    
    recursive_return_dist(h, 10)

def test_density_name_consistency():

    x = cuqi.distribution.Gaussian(geometry=1)
    x2 = x(mean=1)
    x3 = x2(cov=1)

    # Names should be the same as the original density.  
    assert x3.name == 'x'
    assert x2.name == 'x' 
    assert x.name == 'x'

    # Ensure that the name cannot be changed for conditioned densities.
    with pytest.raises(ValueError, match=r"Cannot set name of conditioned density. Only the original density can have its name set."):
        x2.name = 'y'

    x.name = 'y'

    # Ensure that the name is changed for the other conditioned densities.
    assert x2.name == 'y'
    assert x3.name == 'y'

def test_evaluated_density_gradient():
    """ Test that the gradient of the evaluated density is not implemented. """
    x = cuqi.distribution.Gaussian(np.zeros(2), np.eye(2))
    x = x(np.zeros(2)+.1)
    with pytest.raises(NotImplementedError, match=r"gradient is not implemented"):
        x.gradient()



import numpy as np
import scipy.stats as sps
from cuqi.geometry import _get_identity_geometries
from cuqi.utilities import force_ndarray
from cuqi.distribution import Distribution

class InverseGamma(Distribution):
    """
    Multivariate inverse gamma distribution of independent random variables x_i. Each is distributed according to the PDF function

    f(x) = (x-location)^(-shape-1) * exp(-scale/(x-location)) / (scale^(-shape)*Gamma(shape))

    where shape, location and scale are the shape, location and scale of x_i, respectively. And Gamma is the Gamma function.

    Parameters
    ------------
    shape: float or array_like
        The shape parameter

    location: float or array_like
        The location of the inverse gamma distribution. The support of the pdf function is the Cartesian product of the open intervals (location_1, infinity), (location_2, infinity), ..., (location_dim, infinity).

    scale: float or array_like
        The scale of the inverse gamma distribution (non-negative)

    
    Methods
    -----------
    sample: generate one or more random samples
    pdf: evaluate probability density function
    logpdf: evaluate log probability density function
    cdf: evaluate cumulative probability function
    
    Example
    -------
    .. code-block:: python

        # Generate an InverseGamma distribution
        import numpy as np
        import cuqi
        import matplotlib.pyplot as plt
        shape = [1,2]
        location = 0
        scale = 1
        rng = np.random.RandomState(1)
        x = cuqi.distribution.InverseGamma(shape, location, scale)
        samples = x.sample(1000, rng=rng)
        samples.hist_chain(0, bins=70)
        plt.figure()
        samples.hist_chain(1, bins=70)

    """
    def __init__(self, shape=None, location=None, scale=None, is_symmetric=False, **kwargs):
        super().__init__(is_symmetric=is_symmetric, **kwargs) 
        self.shape = force_ndarray(shape, flatten=True)
        self.location = force_ndarray(location, flatten=True)
        self.scale = force_ndarray(scale, flatten=True)
    
    def logpdf(self, x):
        return np.sum(sps.invgamma.logpdf(x, a=self.shape, loc=self.location, scale=self.scale))

    def cdf(self, x):
        return np.prod(sps.invgamma.cdf(x, a=self.shape, loc=self.location, scale=self.scale))

    def _gradient(self, val, **kwargs):
        #Avoid complicated geometries that change the gradient.
        if not type(self.geometry) in _get_identity_geometries():
            raise NotImplementedError("Gradient not implemented for distribution {} with geometry {}".format(self,self.geometry))
        #Computing the gradient for conditional InverseGamma distribution is not supported yet    
        elif self.is_cond:
            raise NotImplementedError(f"Gradient is not implemented for {self} with conditioning variables {self.get_conditioning_variables()}")
        
        #Compute the gradient
        if np.any(val <= self.location):
            return val*np.nan
        else:
            return (-self.shape-1)/(val - self.location) +\
                    self.scale/(val - self.location)**2


    def _sample(self, N=1, rng=None):
        return sps.invgamma.rvs(a=self.shape, loc= self.location, scale = self.scale ,size=(N,self.dim), random_state=rng).T



import sys
from itertools import count
import re

import warnings
import cuqi


def _get_python_variable_name(var):
    """ Retrieve the Python variable name of an object. Takes the first variable name appearing on the stack that is not in the ignore list. """

    ignored_var_names = ["self", "cls", "obj", "var", "_"]

    # First get the stack size and loop (in reverse) through the stack
    # It can be a bit slow to loop through stack size so we limit the levels
    stack_size = _stack_size2a() 

    for i in reversed(range(min(stack_size, cuqi.config.MAX_STACK_SEARCH_DEPTH))): 

        # For each frame we look for a variable name matching the object (var)
        local_vars = sys._getframe(i).f_locals.items()
        var_names = [var_name for var_name, var_val in local_vars if var_val is var]

        # If we find a matching variable name we return it if it is not in the ignore list
        if len(var_names) > 0:
            # Get variable names not regex matching the ignore list
            var_names = [var_name for var_name in var_names if not any([re.match(ignore_var_name, var_name) for ignore_var_name in ignored_var_names])]
            
            if len(var_names) > 0:
                return var_names[0]

    warnings.warn("Could not automatically find variable name for object: {}. Use keyword `name` when defining distribution to specify a name. If code runs slowly and variable name is not needed set config.MAX_STACK_SEARCH_DEPTH to 0.".format(var))

    return None

#https://stackoverflow.com/questions/34115298/how-do-i-get-the-current-depth-of-the-python-interpreter-stack
def _stack_size2a(size=2):
    """Get stack size for caller's frame.
    """
    frame = sys._getframe(size)

    for size in count(size):
        frame = frame.f_back
        if not frame:
            return size



import numpy as np
import scipy.stats as sps
from cuqi.distribution import Distribution
from cuqi.utilities import force_ndarray

class Gamma(Distribution):
    """
    Represents a multivariate Gamma distribution characterized by shape and rate parameters of independent random variables x_i. Each is distributed according to the PDF function
    
    f(x_i; shape, rate) = rate^shape * x_i^(shape-1) * exp(-rate * x_i) / Gamma(shape)

    where `shape` and `rate` are the parameters of the distribution, and Gamma is the Gamma function.

    In case shape and/or rate are arrays, the pdf looks like

    f(x_i; shape_i, rate_i) = rate_i^shape_i * x_i^(shape_i-1) * exp(-rate_i * x_i) / Gamma(shape_i)

    Parameters
    ----------
    shape : float or array_like, optional
        The shape parameter of the Gamma distribution. Must be positive.

    rate : float or array_like, optional
        The rate parameter of the Gamma distribution. Must be positive.

    Examples
    --------
    .. code-block:: python

        import numpy as np
        import cuqi
        import matplotlib.pyplot as plt

        # Create a multivariate Gamma distribution with the same shape and rate parameters
        shape = 1
        rate = 1e-4
        gamma_dist = cuqi.distribution.Gamma(shape=shape, rate=rate, geometry=10)

        # Generate samples
        samples = gamma_dist.sample(10000)

        # Plot histogram of samples for index 0
        samples.hist_chain(0, bins=70)


    .. code-block:: python

        import numpy as np
        import cuqi
        import matplotlib.pyplot as plt

        # Create a multivariate Gamma distribution with different shape and rate parameters
        shape = [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        rate = [1e-4, 1e-3, 1e-2, 1e-1, 1, 1e1, 1e2, 1e3, 1e4, 1e5]
        gamma_dist = cuqi.distribution.Gamma(shape=shape, rate=rate)

        # Generate samples
        samples = gamma_dist.sample(10000)

        # Plot histogram of samples for index 0
        samples.hist_chain(0, bins=70)

    """
    def __init__(self, shape=None, rate=None, is_symmetric=False, **kwargs):
        # Init from abstract distribution class
        super().__init__(is_symmetric=is_symmetric,**kwargs) 

        self.shape = shape
        self.rate = rate

    @property
    def shape(self):
        """ Shape parameter of the Gamma distribution. Must be positive. """
        return self._shape
    
    @shape.setter
    def shape(self, value):
        self._shape = force_ndarray(value, flatten=True)

    @property
    def rate(self):
        """ Rate parameter of the Gamma distribution. Must be positive. """
        return self._rate
    
    @rate.setter
    def rate(self, value):
        self._rate = force_ndarray(value, flatten=True)

    @property
    def scale(self):
        """ Scale parameter of the Gamma distribution. Must be positive. This is the inverse of the rate parameter. """
        return 1/self.rate

    def logpdf(self, x):
        return np.sum(sps.gamma.logpdf(x, a=self.shape, loc=0, scale=self.scale))

    def cdf(self, x):
        return np.prod(sps.gamma.cdf(x, a=self.shape, loc=0, scale=self.scale))

    def _sample(self, N, rng=None):
        if rng is not None:
            return rng.gamma(shape=self.shape, scale=self.scale, size=(N, self.dim)).T
        else:
            return np.random.gamma(shape=self.shape, scale=self.scale, size=(N, self.dim)).T
            


from cuqi.distribution import Posterior, Gaussian, Gamma, GMRF
from cuqi.implicitprior import RegularizedGaussian, RegularizedGMRF
import numpy as np

class Conjugate: # TODO: Subclass from Sampler once updated
    """ Conjugate sampler

    Sampler for sampling a posterior distribution where the likelihood and prior are conjugate.

    Currently supported conjugate pairs are:
    - (Gaussian, Gamma)
    - (GMRF, Gamma)
    - (RegularizedGaussian, Gamma) with nonnegativity constraints only

    For more information on conjugate pairs, see https://en.wikipedia.org/wiki/Conjugate_prior.

    For implicit regularized Gaussians see:
    
    [1] Everink, Jasper M., Yiqiu Dong, and Martin S. Andersen. "Bayesian inference with projected densities." SIAM/ASA Journal on Uncertainty Quantification 11.3 (2023): 1025-1043.

    """

    def __init__(self, target: Posterior):
        if not isinstance(target.likelihood.distribution, (Gaussian, GMRF, RegularizedGaussian, RegularizedGMRF)):
            raise ValueError("Conjugate sampler only works with a Gaussian-type likelihood function")
        if not isinstance(target.prior, Gamma):
            raise ValueError("Conjugate sampler only works with Gamma prior")
        if not target.prior.dim == 1:
            raise ValueError("Conjugate sampler only works with univariate Gamma prior")
            
        if isinstance(target.likelihood.distribution, (RegularizedGaussian, RegularizedGMRF)) and target.likelihood.distribution.preset not in ["nonnegativity"]:
               raise ValueError("Conjugate sampler only works implicit regularized Gaussian likelihood with nonnegativity constraints")
        
        self.target = target

    def step(self, x=None):
        # Extract variables
        b = self.target.likelihood.data                                 #mu
        m = self._calc_m_for_Gaussians(b)                               #n
        Ax = self.target.likelihood.distribution.mean                   #x_i
        L = self.target.likelihood.distribution(np.array([1])).sqrtprec #L
        alpha = self.target.prior.shape                                 #alpha
        beta = self.target.prior.rate                                   #beta

        # Create Gamma distribution and sample
        dist = Gamma(shape=m/2+alpha,rate=.5*np.linalg.norm(L@(Ax-b))**2+beta)

        return dist.sample()

    def _calc_m_for_Gaussians(self, b):
        """ Helper method to calculate m parameter for Gaussian-Gamma conjugate pair. """
        if isinstance(self.target.likelihood.distribution, (Gaussian, GMRF)):
            return len(b)
        elif isinstance(self.target.likelihood.distribution, (RegularizedGaussian, RegularizedGMRF)):
            return np.count_nonzero(b) # See 



from . import data
from . import density
from . import diagnostics
from . import distribution
from . import geometry
from . import implicitprior
from . import likelihood
from . import model
from . import operator
from . import pde
from . import problem
from . import sampler
from . import array
from . import samples
from . import solver
from . import testproblem
from . import utilities

from . import _version
__version__ = _version.get_versions()['version']
